# Copyright 2024 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Commands for HMC user pattern management.
"""


import click

import zhmcclient
from .zhmccli import cli
from ._helper import print_properties, print_resources, abort_if_false, \
    options_to_properties, original_options, COMMAND_OPTIONS_METAVAR, \
    click_exception, add_options, LIST_OPTIONS, FILTER_OPTIONS, \
    build_filter_args, SORT_OPTIONS, build_sort_props, parse_yaml_flow_style


# Special default for retention-time when creating
DEFAULT_RETENTION_TIME = 1


def find_user_pattern(cmd_ctx, console, user_pattern_name):
    """
    Find a user pattern by name and return its resource object.
    """
    try:
        user_pattern = console.user_patterns.find(name=user_pattern_name)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)
    return user_pattern


@cli.group('userpattern', options_metavar=COMMAND_OPTIONS_METAVAR)
def user_pattern_group():
    """
    Command group for managing HMC user patterns.

    User patterns are used when an LDAP user logs on to the HMC. The userid is
    matched against all user patterns, and the matching user pattern identifies
    a user template (i.e. a User object of type template) in one of several
    ways, that is then used to create a pattern-based User object for the
    userid that logs on.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """


@user_pattern_group.command('list', options_metavar=COMMAND_OPTIONS_METAVAR)
@add_options(LIST_OPTIONS)
@add_options(FILTER_OPTIONS)
@add_options(SORT_OPTIONS)
@click.pass_obj
def user_pattern_list(cmd_ctx, **options):
    """
    List the HMC user patterns.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_user_pattern_list(cmd_ctx, options))


@user_pattern_group.command('show', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('USER_PATTERN', type=str, metavar='USER_PATTERN')
@click.pass_obj
def user_pattern_show(cmd_ctx, user_pattern):
    """
    Show the details of an HMC user pattern.

    The following properties are shown in addition to those returned by the HMC:

    \b
      - 'parent-name' - Name of the parent Console.
      - 'specific-template-name' -
        Name of the user template referenced by
        'specific-template-uri'.
      - 'template-name-override-ldap-server-definition-name' -
        Name of the LDAP Server Definition referenced by
        'template-name-override-ldap-server-definition-uri'.
      - 'template-name-override-default-template-name' -
        Name of the user template referenced by
        'template-name-override-default-template-uri'.
      - 'ldap-group-ldap-server-definition-name' -
        Name of the LDAP Server Definition referenced by
        'ldap-group-ldap-server-definition-uri'.
      - 'ldap-group-default-template-name' -
        Name of the user template referenced by
        'ldap-group-default-template-uri'.
      - 'domain-name-restrictions-ldap-server-definition-name' -
        Name of the LDAP Server Definition referenced by
        'domain-name-restrictions-ldap-server-definition-uri'.

    The following old properties are shown that are set by the HMC to be
    consistent with the preferred new properties:

    \b
      - 'ldap-server-definition-name' -
        Name of the LDAP Server Definition referenced by
        'ldap-server-definition-uri'.
      - 'user-template-name' -
        Name of the specific user template referenced by
        'user-template-uri'.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_user_pattern_show(cmd_ctx, user_pattern))


@user_pattern_group.command('create', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.option('--name', type=str, required=True,
              help='The name of the new user pattern.')
@click.option('--description', type=str, required=False,
              help='The description of the new user pattern.')
@click.option('--pattern', type=str, required=True,
              help='The pattern for matching userids. '
              'The combination of "pattern" and "type" must be unique.')
@click.option('--type', type=click.Choice(['glob-like', 'regular-expression']),
              required=True,
              help='The style in which the pattern is expressed.')
@click.option('--retention-time', type=int, required=False,
              default=DEFAULT_RETENTION_TIME,
              help='The time in days that the pattern-based users created '
              'based on this user pattern will be retained after being '
              'created. '
              'A value of 0 indicates not to retain the users after being '
              'created. '
              'Default: {}'.format(DEFAULT_RETENTION_TIME))
@click.option('--specific-template', type=str, required=False,
              # HMC property: 'specific-template-uri/name'
              help='Direct template identification: '
              'The name of the user template to be used. '
              'Default: This template identification method is not used. '
              'HMC property: specific-template-uri/name')
@click.option('--ldap-template-lsd', type=str, required=False,
              # HMC property: 'template-name-override-ldap-server-definition-
              #   uri/name'
              help='LDAP attribute lookup based template identification: '
              'The name of the LDAP Server Definition object identifying '
              'the LDAP server used for lookup of the attribute that specifies '
              'the name of the user template to be used. '
              'Default: This template identification method is not used. '
              'HMC property: template-name-override-ldap-server-'
              'definition-uri/name')
@click.option('--ldap-template-attr', type=str, required=False,
              # HMC property: 'template-name-override'
              help='LDAP attribute lookup based template identification: '
              'The name of the LDAP attribute on the LDAP server that '
              'specifies the name of the user template to be used. '
              'Default: This template identification method is not used. '
              'HMC property: template-name-override')
@click.option('--ldap-template-default', type=str, required=False,
              # HMC property: 'template-name-override-default-template-uri/name'
              help='LDAP attribute lookup based template identification: '
              'The name of the user template to be used when the LDAP '
              'server did not yield a template name. '
              'If not specified, the logon will fail in that case. '
              'Default: This template identification method is not used. '
              'HMC property: template-name-override-default-template-uri/name')
@click.option('--ldap-group-template-lsd', type=str, required=False,
              # HMC property: 'ldap-group-ldap-server-definition-uri/name'
              help='LDAP group lookup based template identification: '
              'The name of the LDAP Server Definition object identifying '
              'the LDAP server used for LDAP group lookup of the user '
              'template name. '
              'Default: This template identification method is not used. '
              'HMC property: ldap-group-ldap-server-definition-uri/name')
@click.option('--ldap-group-template-mappings', type=str, required=False,
              # HMC property: 'ldap-group-to-template-mappings'
              help='LDAP group lookup based template identification: '
              'The mappings from LDAP group names to HMC user '
              'template names, as a dict in YAML Flow Collection style. '
              'The dict key is the LDAP group name, and the dict value is the '
              'name of the template user. '
              'The specified mapping fully replaces the existing mapping in '
              'the HMC. '
              'Default: This template identification method is not used. '
              'HMC property: ldap-group-to-template-mappings. '
              'Example: --ldap-group-template-mappings "{ldap-group-1: '
              'hmc-user-1}"')
@click.option('--ldap-group-template-default', type=str, required=False,
              # HMC property: 'ldap-group-default-template-uri/name'
              help='LDAP group lookup based template identification: '
              'The name of the user template to be used when the LDAP '
              'server does not yield a template name. '
              'If not specified, the logon will fail in that case. '
              'Default: This template identification method is not used. '
              'HMC property: ldap-group-default-template-uri/name')
@click.option('--domain-name-restrictions-lsd', type=str, required=False,
              # HMC property: 'domain-name-restrictions-ldap-server-definition-
              #   uri/name'
              help='Name of the LDAP Server Definition object that '
              'identifies the LDAP server that is used for lookup of the '
              'domain name restrictions. '
              'Default: There are no such restrictions. '
              'HMC property: domain-name-restrictions-ldap-server-'
              'definition-uri/name')
@click.option('--domain-name-restrictions-attr', type=str, required=False,
              # HMC property: 'domain-name-restrictions'
              help='Name of the LDAP attribute on the LDAP server that '
              'is used for lookup of the domain name restrictions, which '
              'specifies the domain name restrictions (i.e. to which consoles '
              'the pattern-based user may log on to). '
              'Default: There are no such restrictions. '
              'HMC property: domain-name-restrictions')
@click.pass_obj
def user_pattern_create(cmd_ctx, **options):
    """
    Create an HMC user pattern.

    The user template to be used when creating the pattern-based user for
    the LDAP user that logs on must be identified using one of these methods.
    The properties related to the methods that are not used, will be reset:

    \b
    * directly by specifying a user template name (--specific-template option)
    * via LDAP attribute lookup (--ldap-template-... options)
    * via LDAP group lookup (--ldap-group-template-... options)

    For some options, the corresponding HMC properties are different than
    the option names. In such cases, the option descriptions mention the
    corresponding HMC property names.

    There are some old properties that are set by the HMC consistent with
    the preferred new properties. Because of that, the old properties have
    no corresponding options and the options for the new preferred properties
    should be used instead. The old properties are:

    \b
    * ldap-server-definition-uri:
      Name of LDAP Server Definition object used for LDAP attribute lookup for
      specific user template name and domain name restrictions. Use the
      --ldap-template-lsd and --domain-name-restrictions-lsd options instead.
    * user-template-uri:
      Name of the specific user template to be used. Use the
      --specific-template option instead.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_user_pattern_create(cmd_ctx, options))


@user_pattern_group.command('update', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('USER_PATTERN', type=str, metavar='USER_PATTERN')
@click.option('--name', type=str, required=False,
              help='The new name of the user pattern.')
@click.option('--description', type=str, required=False,
              help='The new description of the user pattern.')
@click.option('--pattern', type=str, required=False,
              help='The new pattern for matching userids. '
              'The combination of "pattern" and "type" must be unique.')
@click.option('--type', type=click.Choice(['glob-like', 'regular-expression']),
              required=False,
              help='The new style in which the pattern is expressed.')
@click.option('--retention-time', type=int, required=False,
              help='The new time in days that the pattern-based users created '
              'based on this user pattern will be retained after being '
              'created. '
              'A value of 0 indicates not to retain the users after being '
              'created.')
@click.option('--specific-template', type=str, required=False,
              # HMC property: 'specific-template-uri/name'
              help='Direct template identification: '
              'The new name of the user template to be used. '
              'The empty string disables this template identification method. '
              'HMC property: specific-template-uri/name')
@click.option('--ldap-template-lsd', type=str, required=False,
              # HMC property: 'template-name-override-ldap-server-definition-
              #   uri/name'
              help='LDAP attribute lookup based template identification: '
              'The new name of the LDAP Server Definition object identifying '
              'the LDAP server used for lookup of the attribute that specifies '
              'the name of the user template to be used. '
              'The empty string disables this template identification method. '
              'HMC property: template-name-override-ldap-server-'
              'definition-uri/name')
@click.option('--ldap-template-attr', type=str, required=False,
              # HMC property: 'template-name-override'
              help='LDAP attribute lookup based template identification: '
              'The new name of the LDAP attribute on the LDAP server that '
              'specifies the name of the user template to be used. '
              'The empty string disables this template identification method. '
              'HMC property: template-name-override')
@click.option('--ldap-template-default', type=str, required=False,
              # HMC property: 'template-name-override-default-template-uri/name'
              help='LDAP attribute lookup based template identification: '
              'The new name of the user template to be used when the LDAP '
              'server did not yield a template name. '
              'If not specified, the logon will fail in that case. '
              'The empty string disables this template identification method. '
              'HMC property: template-name-override-default-template-uri/name')
@click.option('--ldap-group-template-lsd', type=str, required=False,
              # HMC property: 'ldap-group-ldap-server-definition-uri/name'
              help='LDAP group lookup based template identification: '
              'The new name of the LDAP Server Definition object identifying '
              'the LDAP server used for LDAP group lookup of the user '
              'template name. '
              'The empty string disables this template identification method. '
              'HMC property: ldap-group-ldap-server-definition-uri/name')
@click.option('--ldap-group-template-mappings', type=str, required=False,
              # HMC property: 'ldap-group-to-template-mappings'
              help='LDAP group lookup based template identification: '
              'The new mappings from LDAP group names to HMC user '
              'template names, as a dict in YAML Flow Collection style. '
              'The dict key is the LDAP group name, and the dict value is the '
              'name of the template user. '
              'The specified mapping fully replaces the existing mapping in '
              'the HMC. '
              'The empty string disables this template identification method. '
              'HMC property: ldap-group-to-template-mappings. '
              'Example: --ldap-group-template-mappings "{ldap-group-1: '
              'hmc-user-1}"')
@click.option('--ldap-group-template-default', type=str, required=False,
              # HMC property: 'ldap-group-default-template-uri/name'
              help='LDAP group lookup based template identification: '
              'The new name of the user template to be used when the LDAP '
              'server does not yield a template name. '
              'If not specified, the logon will fail in that case. '
              'The empty string disables this template identification method. '
              'HMC property: ldap-group-default-template-uri/name')
@click.option('--domain-name-restrictions-lsd', type=str, required=False,
              # HMC property: 'domain-name-restrictions-ldap-server-definition-
              #   uri/name'
              help='The new name of the LDAP Server Definition object that '
              'identifies the LDAP server that is used for lookup of the '
              'domain name restrictions. '
              'The empty string sets no such restrictions. '
              'HMC property: domain-name-restrictions-ldap-server-'
              'definition-uri/name')
@click.option('--domain-name-restrictions-attr', type=str, required=False,
              # HMC property: 'domain-name-restrictions'
              help='The new name of the LDAP attribute on the LDAP server that '
              'is used for lookup of the domain name restrictions, which '
              'specifies the domain name restrictions (i.e. to which consoles '
              'the pattern-based user may log on to). '
              'The empty string sets no such restrictions. '
              'HMC property: domain-name-restrictions')
@click.pass_obj
def user_pattern_update(cmd_ctx, user_pattern, **options):
    """
    Update the properties of an HMC user pattern.

    Only the properties will be changed for which a corresponding option is
    specified, so the default for all options is not to change properties.

    The user template to be used when creating the pattern-based user for
    the LDAP user that logs on must be identified using one of these methods.
    The properties related to the methods that are not used, will be reset:

    \b
    * directly by specifying a user template name (--specific-template option)
    * via LDAP attribute lookup (--ldap-template-... options)
    * via LDAP group lookup (--ldap-group-template-... options)

    For some options, the corresponding HMC properties are different than
    the option names. In such cases, the option descriptions mention the
    corresponding HMC property names.

    There are some old properties that are set by the HMC consistent with
    the preferred new properties. Because of that, the old properties have
    no corresponding options and the options for the new preferred properties
    should be used instead. The old properties are:

    \b
    * ldap-server-definition-uri:
      Name of LDAP Server Definition object used for LDAP attribute lookup for
      specific user template name and domain name restrictions. Use the
      --ldap-template-lsd and --domain-name-restrictions-lsd options instead.
    * user-template-uri:
      Name of the specific user template to be used. Use the
      --specific-template option instead.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_user_pattern_update(
        cmd_ctx, user_pattern, options))


@user_pattern_group.command('delete', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('USER_PATTERN', type=str, metavar='USER_PATTERN')
@click.option('-y', '--yes', is_flag=True, callback=abort_if_false,
              expose_value=False,
              help='Skip prompt to confirm deletion of the user pattern.',
              prompt='Are you sure you want to delete this user pattern ?')
@click.pass_obj
def user_pattern_delete(cmd_ctx, user_pattern):
    """
    Delete an HMC user pattern.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_user_pattern_delete(
        cmd_ctx, user_pattern))


def cmd_user_pattern_list(cmd_ctx, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console

    show_list = [
        'name',
    ]
    if not options['names_only']:
        show_list.extend([
            'type',
            'pattern',
            'description',
        ])
    if options['uri']:
        show_list.extend([
            'element-uri',
        ])

    additions = {}

    filter_args = build_filter_args(cmd_ctx, options['filter'])
    try:
        user_patterns = console.user_patterns.list(
            full_properties=False, filter_args=filter_args)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    sort_props = build_sort_props(cmd_ctx, options['sort'], default=['name'])
    try:
        print_resources(cmd_ctx, user_patterns, cmd_ctx.output_format,
                        show_list, additions, all=options['all'],
                        sort_props=sort_props)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)


def cmd_user_pattern_show(cmd_ctx, user_pattern_name):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console
    user_pattern = find_user_pattern(cmd_ctx, console, user_pattern_name)

    try:
        user_pattern.pull_full_properties()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    properties = dict(user_pattern.properties)

    users = console.users.list(filter_args={'type': 'template'})
    lsds = console.ldap_server_definitions.list()

    def _add_user_prop(users, properties, stem):
        uri = properties[stem + 'uri']
        name_prop = stem + 'name'
        if uri:
            objs = [user for user in users if user.uri == uri]
            assert len(objs) == 1
            obj = objs[0]
            properties[name_prop] = obj.name
        else:
            properties[name_prop] = None

    def _add_lsd_prop(lsds, properties, stem):
        uri = properties[stem + 'uri']
        name_prop = stem + 'name'
        if uri:
            objs = [lsd for lsd in lsds if lsd.uri == uri]
            assert len(objs) == 1
            obj = objs[0]
            properties[name_prop] = obj.name
        else:
            properties[name_prop] = None

    properties['parent-name'] = console.name
    _add_user_prop(users, properties, 'specific-template-')
    _add_lsd_prop(lsds, properties,
                  'template-name-override-ldap-server-definition-')
    _add_user_prop(users, properties,
                   'template-name-override-default-template-')
    _add_lsd_prop(lsds, properties, 'ldap-group-ldap-server-definition-')
    _add_user_prop(users, properties, 'ldap-group-default-template-')
    _add_lsd_prop(lsds, properties,
                  'domain-name-restrictions-ldap-server-definition-')
    _add_lsd_prop(lsds, properties, 'ldap-server-definition-')
    _add_user_prop(users, properties, 'user-template-')

    print_properties(cmd_ctx, properties, cmd_ctx.output_format)


def artificial_properties_name_map():
    """
    Return name_map for options_to_properties() for artificial properties.
    """
    return {
        'specific-template': None,
        'ldap-template-lsd': None,
        'ldap-template-attr': 'template-name-override',
        'ldap-template-default': None,
        'ldap-group-template-lsd': None,
        'ldap-group-template-mappings': None,
        'ldap-group-template-default': None,
        'domain-name-restrictions-lsd': None,
        'domain-name-restrictions-attr': 'domain-name-restrictions',
        'old-ldap-lsd': None,
        'old-user-template': None,
    }


def _add_special_name_option(
        cmd_ctx, mgr, properties, org_options, opt_name, prop_name):
    """
    Add an URI property from a name option.
    """
    if org_options[opt_name] is None:
        pass  # omit -> no change
    elif org_options[opt_name] == '':
        properties[prop_name] = None
    else:
        try:
            user = mgr.find_by_name(org_options[opt_name])
        except zhmcclient.Error as exc:
            raise click_exception(exc, cmd_ctx.error_format)
        properties[prop_name] = user.uri


def add_special_options(cmd_ctx, console, properties, org_options):
    """
    Add properties from options that need special handling.
    """
    _add_special_name_option(
        cmd_ctx, console.users, properties, org_options,
        'specific-template',
        'specific-template-uri')
    _add_special_name_option(
        cmd_ctx, console.ldap_server_definitions, properties, org_options,
        'ldap-template-lsd',
        'template-name-override-ldap-server-definition-uri')
    _add_special_name_option(
        cmd_ctx, console.users, properties, org_options,
        'ldap-template-default',
        'template-name-override-default-template-uri')
    _add_special_name_option(
        cmd_ctx, console.ldap_server_definitions, properties, org_options,
        'ldap-group-template-lsd',
        'ldap-group-ldap-server-definition-uri')
    _add_special_name_option(
        cmd_ctx, console.users, properties, org_options,
        'ldap-group-template-default',
        'ldap-group-default-template-uri')
    _add_special_name_option(
        cmd_ctx, console.ldap_server_definitions, properties, org_options,
        'domain-name-restrictions-lsd',
        'domain-name-restrictions-ldap-server-definition-uri')
    _add_special_name_option(
        cmd_ctx, console.ldap_server_definitions, properties, org_options,
        'old-ldap-lsd',
        'ldap-server-definition-uri')
    _add_special_name_option(
        cmd_ctx, console.users, properties, org_options,
        'old-user-template',
        'user-template-uri')

    opt_name = 'ldap-group-template-mappings'
    _opt_name = '--' + opt_name
    prop_name = 'ldap-group-to-template-mappings'
    opt_value = org_options[opt_name]
    if opt_value is None:
        pass  # omit -> no change
    elif opt_value == '':
        properties[prop_name] = None
    else:
        value = parse_yaml_flow_style(cmd_ctx, _opt_name, opt_value)
        if not isinstance(value, dict):
            raise click_exception(
                "Option {} must specify a dictionary, but specified: {!r} "
                "(parsed: {!r})".
                format(_opt_name, opt_value, value), cmd_ctx.error_format)
        prop_value = []
        for ldap_group, template_name in value.items():
            try:
                template = console.users.find_by_name(template_name)
            except zhmcclient.Error as exc:
                raise click_exception(exc, cmd_ctx.error_format)
            prop_value.append({
                'ldap-group-name': ldap_group,
                'template-uri': template.uri,
            })
        properties[prop_name] = prop_value


def cmd_user_pattern_create(cmd_ctx, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console

    org_options = original_options(options)

    name_map = artificial_properties_name_map()
    properties = options_to_properties(org_options, name_map)
    add_special_options(cmd_ctx, console, properties, org_options)

    try:
        new_user_pattern = console.user_patterns.create(properties)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("New user pattern '{r}' has been created.".
               format(r=new_user_pattern.properties['name']))


def cmd_user_pattern_update(cmd_ctx, user_pattern_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console
    user_pattern = find_user_pattern(cmd_ctx, console, user_pattern_name)

    org_options = original_options(options)
    name_map = artificial_properties_name_map()
    properties = options_to_properties(org_options, name_map)
    add_special_options(cmd_ctx, console, properties, org_options)

    if not properties:
        cmd_ctx.spinner.stop()
        click.echo("No properties specified for updating user pattern '{r}'.".
                   format(r=user_pattern_name))
        return

    try:
        user_pattern.update_properties(properties)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("User pattern '{r}' has been updated.".
               format(r=user_pattern_name))


def cmd_user_pattern_delete(cmd_ctx, user_pattern_name):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console
    user_pattern = find_user_pattern(cmd_ctx, console, user_pattern_name)

    try:
        user_pattern.delete()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("User pattern '{r}' has been deleted.".
               format(r=user_pattern_name))



# Copyright 2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Commands for management of character rules within HMC character rule.
"""


import click

import zhmcclient
from ._helper import print_properties, print_dicts, options_to_properties, \
    original_options, COMMAND_OPTIONS_METAVAR, click_exception
from ._cmd_password_rule import password_rule_group, find_password_rule


@password_rule_group.group('characterrule',
                           options_metavar=COMMAND_OPTIONS_METAVAR)
def character_rule_group():
    """
    Command group for managing character rules within HMC password rules.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """


@character_rule_group.command('list', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('PASSWORD_RULE', type=str, metavar='PASSWORD_RULE')
@click.pass_obj
def character_rule_list(cmd_ctx, password_rule):
    """
    List the character rules within an HMC password rule.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_character_rule_list(cmd_ctx, password_rule))


@character_rule_group.command('show', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('PASSWORD_RULE', type=str, metavar='PASSWORD_RULE')
@click.argument('INDEX', type=str, metavar='INDEX')
@click.pass_obj
def character_rule_show(cmd_ctx, password_rule, index):
    """
    Show the details of a character rule at an index within an HMC password
    rule.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_character_rule_show(
        cmd_ctx, password_rule, index))


@character_rule_group.command('add', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('PASSWORD_RULE', type=str, metavar='PASSWORD_RULE')
@click.option('--before', type=int, required=False,
              help='Add the character rule before the rule at this 0-based '
              'index.')
@click.option('--after', type=int, required=False,
              help='Add the character rule after the rule at this 0-based '
              'index.')
@click.option('--last', is_flag=True, required=False,
              help='Add the character rule after the last rule (default).')
@click.option('--min', type=int, required=True,
              help='The minimum number of characters in this part of the '
              'password.')
@click.option('--max', type=int, required=True,
              help='The maximum number of characters in this part of the '
              'password.')
@click.option('--alphabetic',
              type=click.Choice(['allowed', 'not-allowed', 'required']),
              required=True,
              help='The control for the inclusion of alphabetic characters in '
              'this part of the password.')
@click.option('--numeric',
              type=click.Choice(['allowed', 'not-allowed', 'required']),
              required=True,
              help='The control for the inclusion of numeric characters in '
              'this part of the password.')
@click.option('--special',
              type=click.Choice(['allowed', 'not-allowed', 'required']),
              required=True,
              help='The control for the inclusion of special characters '
              '(!?@#$%&^~_+-*/=\\|.,:`";\'(){}[]<>) '
              'in this part of the password.')
@click.option('--custom', nargs=2, metavar='CHARS CONTROL', multiple=True,
              required=False,
              help='The control for the inclusion of special characters '
              '(!?@#$%&^~_+-*/=\\|.,:`";\'(){}[]<>) '
              'in this part of the password.')
@click.pass_obj
def character_rule_add(cmd_ctx, password_rule, **options):
    """
    Add a character rule before/after an index within an HMC password rule.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_character_rule_add(
        cmd_ctx, password_rule, options))


@character_rule_group.command('replace',
                              options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('PASSWORD_RULE', type=str, metavar='PASSWORD_RULE')
@click.argument('INDEX', type=str, metavar='INDEX')
@click.option('--min', type=int, required=True,
              help='The minimum number of characters in this part of the '
              'password.')
@click.option('--max', type=int, required=True,
              help='The maximum number of characters in this part of the '
              'password.')
@click.option('--alphabetic',
              type=click.Choice(['allowed', 'not-allowed', 'required']),
              required=True,
              help='The control for the inclusion of alphabetic characters in '
              'this part of the password.')
@click.option('--numeric',
              type=click.Choice(['allowed', 'not-allowed', 'required']),
              required=True,
              help='The control for the inclusion of numeric characters in '
              'this part of the password.')
@click.option('--special',
              type=click.Choice(['allowed', 'not-allowed', 'required']),
              required=True,
              help='The control for the inclusion of special characters '
              '(!?@#$%&^~_+-*/=\\|.,:`";\'(){}[]<>) '
              'in this part of the password.')
@click.option('--custom', nargs=2, metavar='CHARS CONTROL', multiple=True,
              required=False,
              help='The control for the inclusion of special characters '
              '(!?@#$%&^~_+-*/=\\|.,:`";\'(){}[]<>) '
              'in this part of the password.')
@click.pass_obj
def character_rule_replace(cmd_ctx, password_rule, index, **options):
    """
    Replace a character rule at an index within an HMC password rule.

    All properties of the character rule will be updated. If the option for
    a property is not specified, it will be updated to its default value.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_character_rule_replace(
        cmd_ctx, password_rule, index, options))


@character_rule_group.command('remove', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('PASSWORD_RULE', type=str, metavar='PASSWORD_RULE')
@click.argument('INDEX', type=str, metavar='INDEX')
@click.pass_obj
def character_rule_remove(cmd_ctx, password_rule, index):
    """
    Remove a character rule at an index within an HMC password rule.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_character_rule_remove(
        cmd_ctx, password_rule, index))


def cmd_character_rule_list(cmd_ctx, password_rule_name):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console

    show_list = [
        'index',
        'min-characters',
        'max-characters',
        'alphabetic',
        'numeric',
        'special',
        'custom-character-sets',
    ]

    additions = {}
    additions['index'] = {}

    password_rule = find_password_rule(cmd_ctx, console, password_rule_name)
    character_rules = password_rule.get_property('character-rules')

    # Add artificial property 'index'
    for i, _ in enumerate(character_rules):
        additions['index'][i] = i

    try:
        print_dicts(cmd_ctx, character_rules, cmd_ctx.output_format,
                    show_list, additions)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)


def cmd_character_rule_show(cmd_ctx, password_rule_name, index):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console

    password_rule = find_password_rule(cmd_ctx, console, password_rule_name)
    character_rules = password_rule.get_property('character-rules')

    try:
        index = int(index)
        character_rule = character_rules[index]
    except (ValueError, IndexError):
        raise click.ClickException(
            "Invalid index {i} for character rules of password rule '{r}'".
            format(i=index, r=password_rule_name))

    properties = dict(character_rule)

    print_properties(cmd_ctx, properties, cmd_ctx.output_format)


def cmd_character_rule_add(cmd_ctx, password_rule_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console

    password_rule = find_password_rule(cmd_ctx, console, password_rule_name)
    character_rules = password_rule.get_property('character-rules')
    size = len(character_rules)

    before = options.pop('before')
    after = options.pop('after')
    last = options.pop('last')
    if (bool(before is not None) + bool(after is not None) + last) > 1:
        raise click.ClickException(
            "Only one of --before, --after and --last can be specified.")

    name_map = {
        'min': 'min-characters',
        'max': 'max-characters',
        'custom': None,
    }

    org_options = original_options(options)
    cr_properties = options_to_properties(org_options, name_map)

    # Add custom-character-sets array
    custom_char_sets = []
    for chars, control in options['custom']:
        custom_char_set = {
            'character-set': chars,
            'inclusion': control,
        }
        custom_char_sets.append(custom_char_set)
    cr_properties['custom-character-sets'] = custom_char_sets

    new_character_rules = list(character_rules)
    if before is not None:
        if before < 0 or before >= max(size, 1):
            raise click.ClickException(
                "Invalid --before index {i} for character rules of password "
                "rule '{r}'".
                format(i=before, r=password_rule_name))
        new_character_rules.insert(before, cr_properties)
    elif after is not None:
        if after < 0 or after >= max(size, 1):
            raise click.ClickException(
                "Invalid --after index {i} for character rules of password "
                "rule '{r}'".
                format(i=after, r=password_rule_name))
        new_character_rules.insert(after + 1, cr_properties)
    else:  # last = default
        new_character_rules.insert(size, cr_properties)

    properties = {'character-rules': new_character_rules}

    try:
        password_rule.update_properties(properties)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    where = 'after' if after is not None else 'before'
    index = after if after is not None else before
    click.echo("Password rule '{r}' has been updated to add character rule "
               "{w} index {i}.".
               format(r=password_rule_name, w=where, i=index))


def cmd_character_rule_replace(cmd_ctx, password_rule_name, index, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console

    password_rule = find_password_rule(cmd_ctx, console, password_rule_name)
    character_rules = password_rule.get_property('character-rules')

    try:
        index = int(index)
        _ = character_rules[index]
    except (ValueError, IndexError):
        raise click.ClickException(
            "Invalid index {i} for character rules of password rule '{r}'".
            format(i=index, r=password_rule_name))

    name_map = {
        'min': 'min-characters',
        'max': 'max-characters',
        'custom': None,
    }

    org_options = original_options(options)
    cr_properties = options_to_properties(org_options, name_map)

    # Add custom-character-sets array
    custom_char_sets = []
    for chars, control in options['custom']:
        custom_char_set = {
            'character-set': chars,
            'inclusion': control,
        }
        custom_char_sets.append(custom_char_set)
    cr_properties['custom-character-sets'] = custom_char_sets

    new_character_rules = list(character_rules)
    new_character_rules[index] = cr_properties

    properties = {'character-rules': new_character_rules}

    try:
        password_rule.update_properties(properties)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("Password rule '{r}' has been updated to replace character rule "
               "at index {i}.".
               format(r=password_rule_name, i=index))


def cmd_character_rule_remove(cmd_ctx, password_rule_name, index):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console

    password_rule = find_password_rule(cmd_ctx, console, password_rule_name)
    character_rules = password_rule.get_property('character-rules')

    try:
        index = int(index)
        _ = character_rules[index]
    except (ValueError, IndexError):
        raise click.ClickException(
            "Invalid index {i} for character rules of password rule '{r}'".
            format(i=index, r=password_rule_name))

    new_character_rules = list(character_rules)
    del new_character_rules[index]

    properties = {'character-rules': new_character_rules}

    try:
        password_rule.update_properties(properties)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("Password rule '{r}' has been updated to remove character rule "
               "at index {i}.".
               format(r=password_rule_name, i=index))



# Copyright 2023 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Commands for unmanaged CPCs.
"""


import click

import zhmcclient
from .zhmccli import cli
from ._helper import print_resources, click_exception, \
    COMMAND_OPTIONS_METAVAR, add_options, FILTER_OPTIONS, build_filter_args, \
    SORT_OPTIONS, build_sort_props


def find_unmanaged_cpc(cmd_ctx, console, cpc_name):
    """
    Find an unmanaged CPC by name and return its resource object.
    """
    try:
        ucpc = console.unmanaged_cpcs.find(name=cpc_name)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)
    return ucpc


@cli.group('unmanaged_cpc', options_metavar=COMMAND_OPTIONS_METAVAR)
def ucpc_group():
    """
    Command group for unmanaged CPCs.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """


@ucpc_group.command('list', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.option('--uri', is_flag=True, required=False,
              help='Add the resource URI to the properties shown')
@add_options(FILTER_OPTIONS)
@add_options(SORT_OPTIONS)
@click.pass_obj
def ucpc_list(cmd_ctx, **options):
    """
    List the unmanaged CPCs of the HMC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_ucpc_list(cmd_ctx, options))


def cmd_ucpc_list(cmd_ctx, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console

    filter_args = build_filter_args(cmd_ctx, options['filter'])
    try:
        ucpcs = console.unmanaged_cpcs.list(filter_args=filter_args)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    show_list = [
        'name',
    ]
    if options['uri']:
        show_list.extend([
            'object-uri',
        ])

    sort_props = build_sort_props(cmd_ctx, options['sort'], default=['name'])
    try:
        print_resources(cmd_ctx, ucpcs, cmd_ctx.output_format, show_list,
                        all=False, sort_props=sort_props)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)



# Copyright 2022 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Commands for the HMC console.
"""


import click
import dateutil.parser

import zhmcclient
from .zhmccli import cli
from ._helper import print_resources, print_properties, print_dicts, \
    print_list, TABLE_FORMATS, hide_property, COMMAND_OPTIONS_METAVAR, \
    click_exception, get_level_str, prompt_ftp_password, \
    convert_ec_mcl_description, parse_timestamp, TIMESTAMP_BEGIN_DEFAULT, \
    TIMESTAMP_END_DEFAULT


def find_console_hwmessage(cmd_ctx, console, message_id):
    """
    Find a Hardware messae for this HMC and return its resource object.
    """
    try:
        message = console.hw_messages.find(**{'element-id': message_id})
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)
    return message


@cli.group('console', options_metavar=COMMAND_OPTIONS_METAVAR)
def console_group():
    """
    Command group for the HMC console.

    These commands always target the (one) console of the targeted HMC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """


@console_group.command('show', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.option('--all', is_flag=True, required=False,
              help='Show all properties. Default: Hide some properties in '
              'table output formats')
@click.pass_obj
def console_show(cmd_ctx, **options):
    """
    Show properties of the console of the targeted HMC.

    \b
    In table output formats, the following properties are hidden by default
    but can be shown by using the --all option:
      - ec-mcl-description
      - network-info

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_console_show(cmd_ctx, options))


@console_group.command('restart', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.option('--force', '-f', is_flag=True, required=False,
              help='Perform the restart regardless of whether users are '
              'logged in to the HMC. Users in this sense are local or remote '
              'GUI users. Disconnected GUI users and users at the WS-API do '
              'not count for this.')
@click.option('--wait', '-w', is_flag=True, required=False,
              help='Wait for the HMC to become available again. Note that '
              'any session IDs have become invalid.')
@click.option('--timeout', '-T', type=int, required=False, default=60,
              help='Timeout (in seconds) when waiting for the HMC to become '
              'available again. Default: 60.')
@click.pass_obj
def console_restart(cmd_ctx, **options):
    """
    Restart the targeted HMC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_console_restart(cmd_ctx, options))


@console_group.command('get-audit-log', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.option('--begin', type=str, required=False,
              help='Begin time for the log, in a format supported by '
              'the python-dateutil package. Default: Earliest available.')
@click.option('--end', type=str, required=False,
              help='End time for the log, in a format supported by '
              'the python-dateutil package. Default: Latest available.')
@click.pass_obj
def get_audit_log(cmd_ctx, **options):
    """
    Get the audit log of the targeted HMC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_get_audit_log(cmd_ctx, options))


@console_group.command('get-security-log',
                       options_metavar=COMMAND_OPTIONS_METAVAR)
@click.option('--begin', type=str, required=False,
              help='Begin time for the log, in a format supported by '
              'the python-dateutil package. Default: Earliest available.')
@click.option('--end', type=str, required=False,
              help='End time for the log, in a format supported by '
              'the python-dateutil package. Default: Latest available.')
@click.pass_obj
def get_security_log(cmd_ctx, **options):
    """
    Get the security log of the targeted HMC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_get_security_log(cmd_ctx, options))


@console_group.command('list-api-features',
                       options_metavar=COMMAND_OPTIONS_METAVAR)
@click.option('--name', type=str, metavar='NAME',
              required=False,
              help='A regular expression used to limit returned objects to '
                   'those that have a matching name field.')
@click.pass_obj
def list_api_features(cmd_ctx, **options):
    """
    Lists the Web Services API features available of the console of the
    targeted HMC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(
        lambda: cmd_list_api_features(cmd_ctx, options))


@console_group.command('list-firmware',
                       options_metavar=COMMAND_OPTIONS_METAVAR)
@click.pass_obj
def console_list_firmware(cmd_ctx, **options):
    """
    Lists the firmware levels on the targeted HMC.

    The firmware levels are listed for each EC stream of the HMC as MCL levels
    for different installation states:

    \b
    * retrieved - latest MCL level that has been retrieved
    * installable-conc - latest MCL level that has been retrieved and is
      concurrently installable
    * activated - latest MCL level that has been installed and activated
    * accepted - latest MCL level that has been accepted (= cannot be removed)
    * removable-conc - latest MCL level that has been installed and activated
      and can be removed concurrently (down to latest accepted)

    The MCL levels '0' and '000' are shown as '-' which means there is no such
    level. If a particular installation state is not available, this is shown
    as 'n/a' (but this should not happen).

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(
        lambda: cmd_console_list_firmware(cmd_ctx, options))


@console_group.command('upgrade', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.option('--bundle-level', '-b', type=str, required=False,
              help="Name of the bundle to be installed on the HMC "
              "(e.g. 'H71'). "
              "Default: When --ftp-host is specified, all code changes on "
              "the FTP server will be installed. Otherwise, all locally "
              "available code changes will be installed.")
@click.option('--backup-location-type', type=str, required=False, default='usb',
              help="Type of backup location for the HMC backup that is "
              "performed: 'ftp': The FTP server that was used for the last "
              "console backup as defined on the 'Configure Backup Settings' "
              "user interface task in the HMC GUI., "
              "'usb': The USB storage device mounted to the HMC.")
@click.option('--accept-firmware', '-a', type=bool, required=False,
              default=True,
              help="Boolean indicating to accept the previous bundle level "
              "before installing the new level. Default: true")
@click.option('--ftp-host', type=str, required=False,
              help="The hostname for the FTP server from which the firmware "
              "will be retrieved. "
              "Default: When --bundle-level is specified, firmware will be "
              "retrieved from the IBM support site. Otherwise, all locally "
              "available code changes will be installed.")
@click.option('--ftp-protocol', type=click.Choice(["sftp", "ftp", "ftps"]),
              required=False, default="sftp",
              help="The protocol to connect to the FTP server, if the firmware "
              "is retrieved from an FTP server. Default: sftp.")
@click.option('--ftp-user', type=str, required=False,
              help="The username for the FTP server login, if the firmware "
              "is retrieved from an FTP server.")
@click.option('--ftp-password', type=str, required=False,
              help="The password for the FTP server login, if the firmware "
              "is retrieved from an FTP server. Specifying a hyphen '-' will "
              "prompt for the password.")
@click.option('--ftp-directory', type=str, required=False,
              help="The path name of the directory on the FTP server with the "
              "firmware files, if the firmware is retrieved from an FTP "
              "server.")
@click.option('--timeout', '-T', type=int, required=False, default=1200,
              help='Timeout (in seconds) when waiting for the HMC upgrade '
              'to be complete. Default: 1200.')
@click.pass_obj
def console_upgrade(cmd_ctx, **options):
    """
    Upgrade the firmware on the targeted HMC to a new bundle level.

    This is done by performing the "Console Single Step Install" operation
    which performs the following steps:

    \b
    * If `accept-firmware` is True, the firmware currently installed on the
      this HMC is accepted. Note that once firmware is accepted, it cannot be
      removed.
    * A backup of the this HMC is performed to the specified backup device.
    * The new firmware for the specified bundle level is retrieved from the IBM
      support site or from an FTP server. If no bundle level is specified, but
      an FTP server, all firmware available on the FTP server is retrieved.
      If no bundle level is specified and no FTP server, the already locally
      available firmware is used and no additional firmware is retrieved.
    * The specified firmware is installed.
    * The newly installed firmware is activated, which includes rebooting this
      HMC.

    Note that it is not possible to downgrade the HMC firmware with this
    operation.

    If the HMC firmware is already at the requested bundle level, nothing is
    changed and the command succeeds.

    For HMCs that run on an HMA that also hosts an SE (e.g. z16 and higher),
    the HMC firmware can only be upgraded if the HMA hosts an alternate SE.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_console_upgrade(cmd_ctx, options))


@console_group.group('hw-message', options_metavar=COMMAND_OPTIONS_METAVAR)
def console_hwmessage_group():
    """
    Command group for managing Hardware Messages for the targeted HMC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """


@console_hwmessage_group.command(
    'list', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.option('--all', is_flag=True, required=False,
              help='Show all properties.')
@click.option('--service-supported', type=bool, required=False, default=None,
              help='Filter that narrows the list of returned messages to '
              'those with the specified service-supported value. Note: '
              'Dependent on the number of messages, this may take a while.')
@click.option('--begin', type=str, metavar='TIMESTAMP', required=False,
              help='Filter that narrows the list of returned messages to '
              'those created on or after the specified point in time. '
              'Valid formats are HMC timestamp values and most known formats '
              'to represent date and time. An omitted timezone defaults to '
              'UTC. Other omitted fields default to their earliest possible '
              'values. Ambiguous 3-integer dates (e.g. 01/05/09) are '
              'interpreted as M/D/Y.')
@click.option('--end', type=str, metavar='TIMESTAMP', required=False,
              help='Filter that narrows the list of returned messages to '
              'those created on or before the specified point in time. '
              'Valid formats are HMC timestamp values and most known formats '
              'to represent date and time. An omitted timezone defaults to '
              'UTC. Other omitted fields default to their latest possible '
              'values. Ambiguous 3-integer dates (e.g. 01/05/09) are '
              'interpreted as M/D/Y.')
@click.pass_obj
def console_hwmessage_list(cmd_ctx, **options):
    """
    List the Hardware Messages for the targeted HMC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_console_hwmessage_list(cmd_ctx, options))


@console_hwmessage_group.command(
    'show', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('message_id', type=str, metavar='MESSAGE_ID')
@click.pass_obj
def console_hwmessage_show(cmd_ctx, message_id):
    """
    Show a Hardware Message for the targeted HMC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_console_hwmessage_show(
        cmd_ctx, message_id))


@console_hwmessage_group.command(
    'delete', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('message_id', type=str, metavar='MESSAGE_ID')
@click.pass_obj
def console_hwmessage_delete(cmd_ctx, message_id):
    """
    Delete a Hardware Message for the targeted HMC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_console_hwmessage_delete(
        cmd_ctx, message_id))


@console_hwmessage_group.command(
    'request-service', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('message_id', type=str, metavar='MESSAGE_ID')
@click.option('--customer-name', type=str, required=False,
              help='Name of the person that can be contacted about the '
              'problem. Optional, default is the customer name registered with '
              'IBM for the machine.')
@click.option('--customer-phone', type=str, required=False,
              help='Telephone number of the person that can be contacted '
              'about the problem. Optional, default is the customer phone '
              'registered with IBM for the machine.')
@click.pass_obj
def console_hwmessage_request_service(cmd_ctx, message_id, **options):
    """
    Request service from IBM for a Hardware Message for the targeted HMC and
    delete the hardware message.

    The hardware message's "service-supported" property must be True.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_console_hwmessage_request_service(
        cmd_ctx, message_id, options))


@console_hwmessage_group.command(
    'get-service-info', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('message_id', type=str, metavar='MESSAGE_ID')
@click.option('--delete', is_flag=True, required=False,
              help='Controls whether the hardware message will be deleted '
              'upon successful completion of the operation.')
@click.pass_obj
def console_hwmessage_get_service_info(cmd_ctx, message_id, **options):
    """
    Get problem information and a telephone number for requesting service from
    IBM for a hardware message for the targeted HMC and optionally delete the
    hardware message.

    The hardware message's "service-supported" property must be True.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_console_hwmessage_get_service_info(
        cmd_ctx, message_id, options))


@console_hwmessage_group.command(
    'decline-service', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('message_id', type=str, metavar='MESSAGE_ID')
@click.pass_obj
def console_hwmessage_decline_service(cmd_ctx, message_id):
    """
    Decline service from IBM for a Hardware Message for the targeted HMC and
    delete the hardware message.

    The hardware message's "service-supported" property must be True.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_console_hwmessage_decline_service(
        cmd_ctx, message_id))


def cmd_console_show(cmd_ctx, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console

    try:
        console.pull_full_properties()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    properties = dict(console.properties)

    # Hide some long or deeply nested properties in table output formats.
    if not options['all'] and cmd_ctx.output_format in TABLE_FORMATS:
        hide_property(properties, 'ec-mcl-description')
        hide_property(properties, 'network-info')

    print_properties(cmd_ctx, properties, cmd_ctx.output_format)


def cmd_console_restart(cmd_ctx, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console

    force = options['force']
    wait = options['wait']
    timeout = options['timeout']

    if wait:
        click.echo("Restarting the HMC and waiting for its availability "
                   "(timeout: {} s)".format(timeout))
    else:
        click.echo("Restarting the HMC")
    try:
        console.restart(force=force, wait_for_available=wait,
                        operation_timeout=timeout)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    if wait:
        click.echo("The HMC has restarted and is available again. Any earlier "
                   "session IDs have become invalid")
    else:
        click.echo("The HMC is currently restarting. Wait for it to be "
                   "available again before issuing any commands. Any earlier "
                   "session IDs will be invalid")


def cmd_get_audit_log(cmd_ctx, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console

    begin_time = options['begin'] or None
    if begin_time:
        begin_time = dateutil.parser.parse(begin_time)
    end_time = options['end'] or None
    if end_time:
        end_time = dateutil.parser.parse(end_time)

    try:
        log_items = console.get_audit_log(begin_time, end_time)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    show_list = [
        'timestamp',
        'event-name',
        'event-id',
        'userid',
        'event-message',
    ]

    ts_additions = {}
    for i, log_item in enumerate(log_items):
        hmc_ts = log_item['event-time']
        dt = zhmcclient.datetime_from_timestamp(hmc_ts)
        ts_additions[i] = str(dt)
    additions = {
        'timestamp': ts_additions,
    }

    cmd_ctx.spinner.stop()
    print_dicts(cmd_ctx, log_items, cmd_ctx.output_format,
                show_list=show_list, additions=additions)


def cmd_get_security_log(cmd_ctx, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console

    begin_time = options['begin'] or None
    if begin_time:
        begin_time = dateutil.parser.parse(begin_time)
    end_time = options['end'] or None
    if end_time:
        end_time = dateutil.parser.parse(end_time)

    try:
        log_items = console.get_security_log(begin_time, end_time)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    show_list = [
        'timestamp',
        'event-name',
        'event-id',
        'userid',
        'event-message',
    ]

    ts_additions = {}
    for i, log_item in enumerate(log_items):
        hmc_ts = log_item['event-time']
        dt = zhmcclient.datetime_from_timestamp(hmc_ts)
        ts_additions[i] = str(dt)
    additions = {
        'timestamp': ts_additions,
    }

    cmd_ctx.spinner.stop()
    print_dicts(cmd_ctx, log_items, cmd_ctx.output_format,
                show_list=show_list, additions=additions)


def cmd_list_api_features(cmd_ctx, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console

    name = options['name']
    try:
        features = console.list_api_features(name)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    print_list(cmd_ctx, features, cmd_ctx.output_format)


def cmd_console_list_firmware(cmd_ctx, options):
    # pylint: disable=missing-function-docstring,unused-argument

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console
    console.pull_properties('ec-mcl-description')
    ec_mcl = console.properties['ec-mcl-description']

    firmware_list = convert_ec_mcl_description(ec_mcl)

    # define order of columns in output table
    show_list = [
        'ec-number',
        'description',
        'retrieved',
        'installable-conc',
        'activated',
        'accepted',
        'removable-conc',
    ]

    print_dicts(cmd_ctx, firmware_list, cmd_ctx.output_format,
                show_list=show_list)


def cmd_console_upgrade(cmd_ctx, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console

    bundle_level = options['bundle_level']
    accept_firmware = options['accept_firmware']
    backup_location_type = options['backup_location_type']
    timeout = options['timeout']

    ftp_host = options['ftp_host']
    ftp_user = options['ftp_user']
    ftp_password = options['ftp_password']
    if ftp_host and ftp_password == '-':  # nosec: B105
        ftp_password = prompt_ftp_password(cmd_ctx, ftp_host, ftp_user)

    ec_mcl = console.prop('ec-mcl-description')
    hmc_bundle_level = ec_mcl.get('bundle-level', None)
    if hmc_bundle_level is None:
        hmc_version = console.prop('version')
        raise click_exception(
            "HMC version {v} does not support firmware upgrade through "
            "the Web Services API".format(v=hmc_version),
            cmd_ctx.error_format)

    level_str = get_level_str(bundle_level, ftp_host)
    click.echo("Upgrading the HMC to {lvl}, and waiting for completion "
               "(timeout: {t} s)".
               format(lvl=level_str, t=timeout))

    kwargs = dict(
        bundle_level=bundle_level,
        accept_firmware=accept_firmware,
        backup_location_type=backup_location_type,
        wait_for_completion=True,
        operation_timeout=timeout)
    if ftp_host:
        kwargs['ftp_host'] = ftp_host
        kwargs['ftp_protocol'] = options['ftp_protocol']
        kwargs['ftp_user'] = ftp_user
        kwargs['ftp_password'] = ftp_password
        kwargs['ftp_directory'] = options['ftp_directory']

    try:
        console.single_step_install(**kwargs)
    except zhmcclient.HTTPError as exc:
        if exc.http_status == 400 and exc.reason == 356:
            # HMC was already at that bundle level
            cmd_ctx.spinner.stop()
            click.echo("The HMC was already at {lvl} and did not need to be "
                       "upgraded".
                       format(lvl=level_str))
            return
        raise click_exception(exc, cmd_ctx.error_format)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("The HMC has been upgraded to {lvl} and is available again. "
               "Any earlier session IDs have become invalid".
               format(lvl=level_str))


def cmd_console_hwmessage_list(cmd_ctx, options):
    # pylint: disable=missing-function-docstring
    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console

    begin_time = options['begin']
    end_time = options['end']
    service_supported = options['service_supported']

    if begin_time is not None:
        try:
            begin_time = parse_timestamp(begin_time, TIMESTAMP_BEGIN_DEFAULT)
        except (ValueError, OverflowError) as exc:
            raise click_exception(
                f"Invalid begin time: {exc}", cmd_ctx.error_format)
    if end_time is not None:
        try:
            end_time = parse_timestamp(end_time, TIMESTAMP_END_DEFAULT)
        except (ValueError, OverflowError) as exc:
            raise click_exception(
                f"Invalid end time: {exc}", cmd_ctx.error_format)

    filter_args = {}
    if service_supported is not None:
        filter_args['service-supported'] = service_supported

    try:
        messages = console.hw_messages.list(
            filter_args=filter_args,
            begin_time=begin_time,
            end_time=end_time,
        )
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    show_list = [
        'timestamp-utc',
        'message-id',
        'text',
    ]

    ts_additions = {}
    msgid_additions = {}
    for message in messages:
        hmc_ts = message.prop('timestamp')
        dt = zhmcclient.datetime_from_timestamp(hmc_ts)
        ts_additions[message.uri] = dt.strftime('%Y-%m-%d %H:%M:%S.%f')
        msgid_additions[message.uri] = message.name
    additions = {
        'timestamp-utc': ts_additions,
        'message-id': msgid_additions,
    }

    try:
        print_resources(cmd_ctx, messages, cmd_ctx.output_format,
                        show_list, additions=additions, all=options['all'])
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)


def cmd_console_hwmessage_show(cmd_ctx, message_id):
    # pylint: disable=missing-function-docstring
    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console
    message = find_console_hwmessage(cmd_ctx, console, message_id)
    message.pull_full_properties()

    message.update_properties_local({'message-id': message.name})
    hmc_ts = message.prop('timestamp')
    dt = zhmcclient.datetime_from_timestamp(hmc_ts)
    dt_str = dt.strftime('%Y-%m-%d %H:%M:%S.%f')
    message.update_properties_local({'timestamp-utc': dt_str})

    cmd_ctx.spinner.stop()
    print_properties(cmd_ctx, message.properties, cmd_ctx.output_format)


def cmd_console_hwmessage_delete(cmd_ctx, message_id):
    # pylint: disable=missing-function-docstring
    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console
    message = find_console_hwmessage(cmd_ctx, console, message_id)

    try:
        message.delete()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo(f"Hardware Message '{message.name}' has been deleted.")


def cmd_console_hwmessage_request_service(cmd_ctx, message_id, options):
    # pylint: disable=missing-function-docstring
    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console
    message = find_console_hwmessage(cmd_ctx, console, message_id)

    try:
        message.request_service(**options)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo(f"IBM service for Hardware Message '{message.name}' has been "
               "requested and the message has been deleted.")


def cmd_console_hwmessage_get_service_info(cmd_ctx, message_id, options):
    # pylint: disable=missing-function-docstring
    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console
    message = find_console_hwmessage(cmd_ctx, console, message_id)
    delete = options['delete']

    try:
        info = message.get_service_information(delete=delete)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    print_properties(cmd_ctx, info, cmd_ctx.output_format)
    if delete:
        click.echo(f"Hardware Message '{message.name}' has been deleted.")


def cmd_console_hwmessage_decline_service(cmd_ctx, message_id):
    # pylint: disable=missing-function-docstring
    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console
    message = find_console_hwmessage(cmd_ctx, console, message_id)

    try:
        message.decline_service()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo(f"IBM service for Hardware Message '{message.name}' has been "
               "declined and the message has been deleted.")



# Copyright 2025 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
End2end tests for the 'zhmc console hw-message' command group.
"""

import re
import random
import urllib3
import pytest
import zhmcclient
# pylint: disable=unused-import
from zhmcclient.testutils import hmc_definition  # noqa: F401
from .utils import zhmc_session  # noqa: F401
# pylint: enable=unused-import

from .utils import run_zhmc, parse_output, fixup

urllib3.disable_warnings()


# pylint: disable=inconsistent-return-statements
def pick_console_with_hwmessages(console):
    """
    Pick the console if it has at least one hardware message.
    """
    messages = console.hw_messages.list()
    if messages:
        return console, messages

    pytest.skip("Did not find a CPC with hardware messages")


def assert_messages(
        rc, stdout, stderr, out_format, command, exp_rc, exp_stderr, exp_props,
        exp_messages):
    """
    Check exit code, stdout, stderr of the zhmc command against
    expected values.

    This function can be used for two cases:
    * command='list': Check a list of hardware messages in stdout, against the
      same set of messages in exp_messages.
    * command='show': Check a single hardware message in stdout against a
      single message in exp_messages (it is a list with one item).
    """
    assert rc == exp_rc, (
        f"Unexpected exit code {rc} (expected was {exp_rc})\n"
        f"Stdout: {stdout}\n"
        f"Stderr: {stderr}\n"
    )

    if command == 'list':
        out_type = list
    else:
        assert command == 'show'
        out_type = dict

    if exp_rc == 0:
        assert stderr == ""

        out_obj = parse_output(stdout, out_format, out_type)

        if command == 'list':
            assert len(out_obj) == len(exp_messages)
            # Check only one message from the result
            index = random.randrange(len(out_obj))  # nosec
            out_item = out_obj[index]
            exp_message = exp_messages[index]
        else:
            out_item = out_obj
            exp_message = exp_messages[0]

        # Check the message
        message_id = exp_message.prop('element-id')
        for pname in exp_props:
            if pname not in CPC_HWMSG_PROPS_ARTIFICIAL:
                assert pname in out_item
                exp_value = exp_message.prop(pname, 'undefined')
                if out_format == 'csv':
                    value = fixup(out_item[pname])
                else:
                    value = out_item[pname]
                assert value == exp_value, (
                    f"Unexpected value for property {pname} of hardware "
                    f"message {message_id} of a CPC:\n"
                    f"Expected: {exp_value!r}\n"
                    f"Actual: {value!r}"
                )
    else:
        assert re.search(exp_stderr, stderr, re.MULTILINE)


# Expected output properties when no other options are specified
CPC_HWMSG_PROPS_DEFAULT = [
    "timestamp-utc",
    "message-id",
    "text",
]

# Expected output properties when --all option is specified
# Note: These are the z16 properties. Additional properties are allowed.
CPC_HWMSG_PROPS_ALL = [
    "timestamp-utc",
    "message-id",
    "text",
    "class",
    "details",
    "element-id",
    "element-uri",
    "parent",
    "service-supported",
    "text",
    "timestamp",
]

# Artificial output properties (that are not properties on the HardwareMessage
# object).
CPC_HWMSG_PROPS_ARTIFICIAL = [
    "timestamp-utc",
    "message-id",
]


TESTCASES_CPC_HWMSG_LIST = [
    # Testcases for test_console_hwmsg_list()
    # Each list item is a testcase, which is a tuple with items:
    # - options (list of str): Options for 'zhmc console hw-message list' cmd
    # - exp_props (list of str): Expected property names in output
    (
        [],
        CPC_HWMSG_PROPS_DEFAULT,
    ),
    (
        ['--all'],
        CPC_HWMSG_PROPS_ALL,
    ),
]


@pytest.mark.parametrize(
    "options, exp_props",
    TESTCASES_CPC_HWMSG_LIST
)
@pytest.mark.parametrize(
    "out_format",
    ['json', 'csv']
)
def test_console_hwmsg_list(
        zhmc_session, out_format, options, exp_props):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test 'console hw-message list' command for output formats json and csv.
    """
    client = zhmcclient.Client(zhmc_session)
    console = client.consoles.console
    console, messages = pick_console_with_hwmessages(console)

    # Reduce messages to be listed since --all takes a long time for many
    # messages
    num_messages = min(5, len(messages))
    messages = messages[-num_messages:]
    first_ts_dt = zhmcclient.datetime_from_timestamp(
        messages[0].prop('timestamp'))
    first_ts_str = first_ts_dt.strftime('%Y-%m-%d %H:%M:%S.%f')

    args = ['-o', out_format, 'console', 'hw-message', 'list', '--begin',
            first_ts_str] + options
    rc, stdout, stderr = run_zhmc(args)

    assert_messages(rc, stdout, stderr, out_format, 'list', 0, "",
                    exp_props, messages)


@pytest.mark.parametrize(
    "out_format",
    ['json', 'csv']
)
def test_console_hwmsg_show(zhmc_session, out_format):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test 'console hw-message show' command.
    """
    client = zhmcclient.Client(zhmc_session)
    console = client.consoles.console
    console, messages = pick_console_with_hwmessages(console)

    test_message = random.choice(messages)  # nosec
    test_message_id = test_message.prop('element-id')

    rc, stdout, stderr = run_zhmc(
        ['-o', out_format, 'console', 'hw-message', 'show', test_message_id])

    assert_messages(rc, stdout, stderr, out_format, 'show', 0, "",
                    CPC_HWMSG_PROPS_DEFAULT, [test_message])


# Note: We do not have end2end tests for modifying commands:
# 'console hw-message delete'
# 'console hw-message get-service-info'
# 'console hw-message request-service'
# 'console hw-message decline-service'



# Copyright 2016,2019 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Commands for CPCs.
"""


from datetime import datetime, timezone
import json

import click
import yaml
import yamlloader
import zhmcclient
from tabulate import tabulate

from ._helper import print_properties, print_resources, print_list, \
    options_to_properties, original_options, COMMAND_OPTIONS_METAVAR, \
    click_exception, add_options, LIST_OPTIONS, FILTER_OPTIONS, \
    build_filter_args, SORT_OPTIONS, build_sort_props, TABLE_FORMATS, \
    hide_property, required_option, validate, print_dicts, get_level_str, \
    prompt_ftp_password, convert_ec_mcl_description, get_mcl_str, \
    parse_ec_levels, parse_timestamp, TIMESTAMP_BEGIN_DEFAULT, \
    TIMESTAMP_END_DEFAULT
from ._version import __version__
from .zhmccli import cli

POWER_SAVING_TYPES = ['high-performance', 'low-power', 'custom']
DEFAULT_POWER_SAVING_TYPE = 'high-performance'
POWER_CAPPING_STATES = ['disabled', 'enabled', 'custom']

# Data formats of DPM configuration file
DPM_FORMATS = ['yaml', 'json']
DEFAULT_DPM_FORMAT = 'yaml'


# JSON schema for adapter mapping file
MAPPING_SCHEMA = {
    "title": "adapter mapping file schema",
    "description":
        "JSON schema that defines the structure of an adapter mapping file.",
    "type": "object",
    "additionalProperties": False,
    "required": [
        "adapter-mapping"
    ],
    "properties": {
        "adapter-mapping": {
            "description":
                "List of PCHID mappings from config file to import into CPC.",
            "type": "array",
            "items": {
                "type": "object",
                "additionalProperties": False,
                "required": [
                    "old-adapter-id",
                    "new-adapter-id",
                ],
                "properties": {
                    "old-adapter-id": {
                        "description":
                            "PCHID of adapter in DPM configuration file.",
                        "type": "string"
                    },
                    "new-adapter-id": {
                        "description":
                            "PCHID of adapter when imported into the CPC.",
                        "type": "string"
                    }
                }
            }
        }
    }
}


def help_dpm_file(cmd_ctx, param, value):
    # pylint: disable=unused-argument
    """
    Click callback function for --help-dpm-file option, that displays
    help for the format of the DPM configuration file and exits.
    """
    if not value or cmd_ctx.resilient_parsing:
        return
    print("""
Format of DPM configuration file:

The DPM configuration file contains the properties of a single CPC and all of
its child objects (such as partitions or adapters) and all of its CPC-specific
related objects (such as storage groups associated with the CPC).

The DPM configuration file is written by the 'zhmc cpc dpm-export' command
and read by the 'zhmc cpc dpm-import' command.

The DPM configuration file is in YAML or JSON format and has the structure of
the payload of the 'Import DPM Configuration' operation described in the
HMC API book.

When including any of following properties, remember that the
'zhmc cpc dpm-import' command may overwrite these properties based on the
corresponding options when issuing the 'Import DPM Configuration' operation:

* 'preserve-uris' - Boolean controlling whether to preserve object URIs and IDs.

* 'preserve-wwpns' - Boolean controlling whether to preserve HBA WWPNs.

* 'adapter-mapping' - List of mappings of adapter PCHIDs between the CPC from
  which the DPM configuration was exported and the new CPC to which it is
  imported.
""")
    cmd_ctx.exit()


def help_mapping_file(cmd_ctx, param, value):
    # pylint: disable=unused-argument
    """
    Click callback function for --help-dpm-file option, that displays
    help for the format of the adapter mapping file and exits.
    """
    if not value or cmd_ctx.resilient_parsing:
        return
    print("""
Format of adapter mapping file:

The adapter mapping file specifies how PCHIDs of adapters in a DPM
configuration file need to be replaced when importing the DPM configuration
into a new CPC. If you do not provide a mapping for an adapter, DPM uses a
one-to-one mapping of adapters in the configuration file to adapters on the
target system.

The adapter mapping file is created manually and is read by the
'zhmc cpc dpm-import' command.

The adapter mapping file is in YAML format and has the following structure:

    adapter-mapping:
      - old-adapter-id:  A11  # PCHID in the DPM configuration file
        new-adapter-id: "911" # PCHID in the new system that maps to A11
      - old-adapter-id: "11c"
        new-adapter-id: "12C"
      - ... # More mappings, one for each adapter PCHID you intend to map

The values of the 'old-adapter-id' and 'new-adapter-id' properties are the
PCHID values as a hexadecimal string. Therefore, hexadecimal values that consist
only of decimal digits (like '911' in the example above) must be put into double
quotes. The characters in the hexadecimal string may be in upper or lower case.
""")
    cmd_ctx.exit()


def find_cpc(cmd_ctx, client, cpc_name):
    """
    Find a CPC by name and return its resource object.
    """
    try:
        cpc = client.cpcs.find(name=cpc_name)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)
    return cpc


def find_cpc_hwmessage(cmd_ctx, cpc, message_id):
    """
    Find a Hardware messae of a CPC and return its resource object.
    """
    try:
        message = cpc.hw_messages.find(**{'element-id': message_id})
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)
    return message


@cli.group('cpc', options_metavar=COMMAND_OPTIONS_METAVAR)
def cpc_group():
    """
    Command group for managing CPCs.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """


@cpc_group.command('list', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.option('--type', is_flag=True, required=False, hidden=True)
@click.option('--mach', is_flag=True, required=False, hidden=True)
@add_options(LIST_OPTIONS)
@add_options(FILTER_OPTIONS)
@add_options(SORT_OPTIONS)
@click.pass_obj
def cpc_list(cmd_ctx, **options):
    """
    List the CPCs managed by the HMC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_cpc_list(cmd_ctx, options))


@cpc_group.command('show', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.option('--all', is_flag=True, required=False,
              help='Show all properties. Default: Hide some properties in '
              'table output formats')
@click.pass_obj
def cpc_show(cmd_ctx, cpc, **options):
    """
    Show details of a CPC.

    \b
    In table output formats, the following properties are hidden by default
    but can be shown by using the --all option:
      - auto-start-list
      - available-features-list
      - cpc-power-saving-state
      - ec-mcl-description
      - network1-ipv6-info
      - network2-ipv6-info
      - stp-configuration

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_cpc_show(cmd_ctx, cpc, options))


@cpc_group.command('update', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.option('--description', type=str, required=False,
              help='The new description of the CPC. '
              '(DPM mode only).')
@click.option('--acceptable-status', type=str, required=False,
              help='The new set of acceptable operational status values, as a '
              'comma-separated list. The empty string specifies an empty list.')
@click.option('--next-activation-profile', type=str, required=False,
              help='The name of the new next reset activation profile. '
              '(not in DPM mode).')
@click.option('--processor-time-slice', type=int, required=False,
              help='The new time slice (in ms) for logical processors. '
              'A value of 0 causes the time slice to be dynamically '
              'determined by the system. A positive value causes a constant '
              'time slice to be used. '
              '(not in DPM mode).')
@click.option('--wait-ends-slice/--no-wait-ends-slice', default=None,
              required=False,
              help='The new setting for making logical processors lose their '
              'time slice when they enter a wait state. '
              '(not in DPM mode).')
@click.pass_obj
def cpc_update(cmd_ctx, cpc, **options):
    """
    Update the properties of a CPC.

    Only the properties will be changed for which a corresponding option is
    specified, so the default for all options is not to change properties.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_cpc_update(cmd_ctx, cpc, options))


@cpc_group.command('set-power-save', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.option('--power-saving', type=click.Choice(POWER_SAVING_TYPES),
              required=False, default=DEFAULT_POWER_SAVING_TYPE,
              help='Defines the type of power saving. Default: {pd}'.
              format(pd=DEFAULT_POWER_SAVING_TYPE))
@click.pass_obj
def set_power_save(cmd_ctx, cpc, **options):
    """
    Set the power save settings of a CPC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_cpc_set_power_save(cmd_ctx, cpc, options))


@cpc_group.command('set-power-capping',
                   options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.option('--power-capping-state', type=click.Choice(POWER_CAPPING_STATES),
              required=True,
              help='Defines the state of power capping.')
@click.option('--power-cap-current', type=int, required=False,
              help='Specifies the current cap value for the CPC in watts (W). '
              'Required if power capping state is enabled.')
@click.pass_obj
def set_power_capping(cmd_ctx, cpc, **options):
    """
    Set the power capping settings of a CPC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_cpc_set_power_capping(cmd_ctx, cpc,
                                                          options))


@cpc_group.command('get-em-data', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.pass_obj
def get_em_data(cmd_ctx, cpc):
    """
    Get all energy management data of a CPC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_cpc_get_em_data(cmd_ctx, cpc))


@cpc_group.command('dpm-export', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC', required=True)
@click.option('--dpm-file', '-d', metavar='FILE', type=str, required=True,
              help='Path name of the DPM configuration file to be written.')
@click.option('--dpm-format', '-f', type=click.Choice(DPM_FORMATS),
              required=False, default=DEFAULT_DPM_FORMAT,
              help='Controls in which data format the DPM configuration file '
              'is written. Default: {d}.'.format(d=DEFAULT_DPM_FORMAT))
@click.option('--exclude-meta-fields', '-e', is_flag=True, default=False,
              help='Controls whether additional meta information should be '
              'added after a successful export operation, regarding details '
              'of the export operation itself: CPC name, exporting user, '
              'timestamp of export. All corresponding fields are named '
              '"zhmccli-meta-xxx". Default: included')
@click.option('--help-dpm-file', is_flag=True, is_eager=True,
              callback=help_dpm_file, expose_value=False,
              help='Show help on the format of the DPM configuration file '
              'and exit.')
@click.option('--include-unused-adapters', '-i', is_flag=True, default=False,
              help='Controls whether the full set of adapters should be '
              'returned, vs. only those that are referenced by other DPM '
              'objects that are part of the return data. Default: not included')
@click.pass_obj
def dpm_export(cmd_ctx, cpc, **options):
    """
    Export a DPM configuration from a CPC.

    The DPM configuration of the CPC is exported and written to a DPM
    configuration file (in YAML or JSON format). Note: the configuration
    file contains additional meta information, use --exclude-meta-fields in
    case you intend to import the configuration file with other tooling than
    zhmc. For details about the format of that file, use --help-dpm-file.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_dpm_export(cmd_ctx, cpc, options))


@cpc_group.command('dpm-import', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC', required=True)
@click.option('-y', '--yes', is_flag=True,
              help='Skip prompt to confirm import of the DPM configuration.')
@click.option('--dpm-file', '-d', metavar='FILE', type=str, required=True,
              help='Path name of the DPM configuration file to be used.')
@click.option('--dpm-format', '-f', type=click.Choice(DPM_FORMATS),
              required=False, default=DEFAULT_DPM_FORMAT,
              help='Controls in which data format the DPM configuration file '
              'is read. Default: {d}.'.format(d=DEFAULT_DPM_FORMAT))
@click.option('--preserve-uris/--generate-uris',
              'preserve_uris', default=None,
              help='Controls whether existing URIs and IDs of objects in the '
              'DPM configuration file are preserved or ignored and new ones '
              'are generated. If either flag is present, zhmc overwrites any '
              'potential corresponding setting within the configuration data.')
@click.option('--preserve-wwpns/--generate-wwpns',
              'preserve_wwpns', default=None,
              help='Controls whether existing WWPNs of HBAs in the DPM '
              'configuration file are preserved or ignored and new ones '
              'are generated. If either flag is present, zhmc overwrites any '
              'potential corresponding setting within the configuration data.')
@click.option('--mapping-file', '-m', metavar='FILE', type=str, required=False,
              help='Path name of the adapter mapping file to be used. If '
              'present, zhmc overwrites any potential corresponding setting '
              'within the configuration data. Default: Use 1:1 adapter '
              'mapping.')
@click.option('--help-dpm-file', is_flag=True, is_eager=True,
              callback=help_dpm_file, expose_value=False,
              help='Show help on the format of the DPM configuration file '
              'and exit.')
@click.option('--help-mapping-file', is_flag=True, is_eager=True,
              callback=help_mapping_file, expose_value=False,
              help='Show help on the format of the adapter mapping file '
              'and exit.')
@click.pass_obj
def dpm_import(cmd_ctx, cpc, **options):
    """
    Import a DPM configuration into a CPC.

    The DPM configuration of the CPC is read from a DPM configuration file
    (in YAML or JSON format) and imported into the CPC, replacing the current
    DPM configuration of the CPC.
    For details about the format of that file, use --help-dpm-file.

    Optionally, an adapter mapping file (in YAML format) can be specified that
    can be used to accommodate different adapter PCHIDs (plug positions) between
    the CPC represented in the DPM configuration and the CPC targeted for the
    import. By default, the adapters are assumed to be in the same plug
    positions, i.e. the PCHIDs will be unchanged.
    For details about the format of that file, use --help-mapping-file.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.

    Importing a configuration can potentially result in significant changes to
    the corresponding CPC. To reduce the risk of importing unwanted data,
    this method is more verbose than usual and summarizes the key elements
    of the configuration that is subject to be imported.
    """
    cmd_ctx.execute_cmd(lambda: cmd_dpm_import(cmd_ctx, cpc, options))


@cpc_group.group('autostart', options_metavar=COMMAND_OPTIONS_METAVAR)
def cpc_autostart_group():
    """
    Command group for managing the auto-start list of a CPC (in DPM mode).

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """


@cpc_autostart_group.command('show', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.pass_obj
def cpc_autostart_show(cmd_ctx, cpc):
    """
    Show the auto-start list of a CPC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_cpc_autostart_show(cmd_ctx, cpc))


@cpc_autostart_group.command('add', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('PARTITIONS_DELAY', type=(str, int), metavar='PARTITIONS DELAY')
@click.option('--group', type=str, metavar='GROUP',
              required=False,
              help='Add the partition(s) as a partition group with this name. '
              'Required when adding a group.')
@click.option('--description', type=str, metavar='TEXT',
              required=False,
              help='Description of partition group. '
              'Default: No description.')
@click.option('--before', type=str, metavar='PARTITION_OR_GROUP',
              required=False,
              help='Insert the new partition or group before this '
              'partition/group. '
              'Default: Append new partition or group to the end.')
@click.option('--after', type=str, metavar='PARTITION_OR_GROUP',
              required=False,
              help='Insert the new partition or group after this '
              'partition/group. '
              'Default: Append new partition or group to the end.')
@click.pass_obj
def cpc_autostart_add(cmd_ctx, cpc, partitions_delay, **options):
    """
    Add a partition or group to the auto-start list of a CPC.

    A partition group exists only in context of the auto-start list; it has
    nothing to do with Group objects.

    PARTITIONS is the partition name or a comma-separated list of partition
    names in case of adding a partition group.

    DELAY is the delay afer starting this partition or group, in seconds.

    The updated auto-start list is shown.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(
        lambda: cmd_cpc_autostart_add(cmd_ctx, cpc, partitions_delay, options))


@cpc_autostart_group.command('delete', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('PARTITION_OR_GROUP', type=str, metavar='PARTITION_OR_GROUP')
@click.pass_obj
def cpc_autostart_delete(cmd_ctx, cpc, partition_or_group):
    """
    Delete a partition or group from the auto-start list of a CPC.

    The updated auto-start list is shown.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(
        lambda: cmd_cpc_autostart_delete(cmd_ctx, cpc, partition_or_group))


@cpc_autostart_group.command('clear', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.pass_obj
def cpc_autostart_clear(cmd_ctx, cpc):
    """
    Clear the auto-start list of a CPC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(
        lambda: cmd_cpc_autostart_clear(cmd_ctx, cpc))


@cpc_group.command('list-api-features', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.option('--name', type=str, metavar='NAME',
              required=False,
              help='A regular expression used to limit returned objects to '
                   'those that have a matching name field.')
@click.pass_obj
def cpc_list_api_features(cmd_ctx, cpc, **options):
    """
    List the Web Services API features available on a CPC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(
        lambda: cmd_cpc_list_api_features(cmd_ctx, cpc, options))


@cpc_group.command('list-firmware', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.pass_obj
def cpc_list_firmware(cmd_ctx, cpc, **options):
    """
    List the firmware level on the Support Element (SE) of a CPC.

    The firmware levels are listed for each EC stream of the SE as MCL levels
    for different installation states:

    \b
    * retrieved - latest MCL level that has been retrieved
    * installable-conc - latest MCL level that has been retrieved and is
      concurrently installable
    * activated - latest MCL level that has been installed and activated
    * accepted - latest MCL level that has been accepted (= cannot be removed)
    * removable-conc - latest MCL level that has been installed and activated
      and can be removed concurrently (down to latest accepted)

    The MCL levels '0' and '000' are shown as '-' which means there is no such
    level. If a particular installation state is not available, this is shown
    as 'n/a' (but this should not happen).

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(
        lambda: cmd_cpc_list_firmware(cmd_ctx, cpc, options))


@cpc_group.command('upgrade', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.option('--bundle-level', '-b', type=str, required=False,
              help="Name of the bundle to be installed on the SE "
              "(e.g. 'S71'). "
              "Default: When --ftp-host is specified, all code changes on "
              "the FTP server will be installed. Otherwise, all locally "
              "available code changes will be installed.")
@click.option('--accept-firmware', '-a', type=bool, required=False,
              default=True,
              help="Boolean indicating to accept the previous bundle level "
              "before installing the new level. Default: true")
@click.option('--ftp-host', type=str, required=False,
              help="The hostname for the FTP server from which the firmware "
              "will be retrieved. "
              "Default: When --bundle-level is specified, firmware will be "
              "retrieved from the IBM support site. Otherwise, all locally "
              "available code changes will be installed.")
@click.option('--ftp-protocol', type=click.Choice(["sftp", "ftp", "ftps"]),
              required=False, default="sftp",
              help="The protocol to connect to the FTP server, if the firmware "
              "is retrieved from an FTP server. Default: sftp.")
@click.option('--ftp-user', type=str, required=False,
              help="The username for the FTP server login, if the firmware "
              "is retrieved from an FTP server.")
@click.option('--ftp-password', type=str, required=False,
              help="The password for the FTP server login, if the firmware "
              "is retrieved from an FTP server. Specifying a hyphen '-' will "
              "prompt for the password.")
@click.option('--ftp-directory', type=str, required=False,
              help="The path name of the directory on the FTP server with the "
              "firmware files, if the firmware is retrieved from an FTP "
              "server.")
@click.option('--accept-firmware', '-a', type=bool, required=False,
              default=True,
              help="Boolean indicating to accept the previous bundle level "
              "before installing the new level. Default: true")
@click.option('--timeout', '-T', type=int, required=False, default=1200,
              help='Timeout (in seconds) when waiting for the SE upgrade '
              'to be complete. Default: 1200.')
@click.pass_obj
def cpc_upgrade(cmd_ctx, cpc, **options):
    """
    Upgrade the firmware in a single step on the Support Element (SE) of a CPC.

    This is done by performing the "CPC Single Step Install" operation
    which performs the following steps:

    \b
    * A backup of the target CPC is performed to its SE hard drive.
    * If `accept_firmware` is True, the firmware currently installed on the SE
      of this CPC is accepted. Note that once firmware is accepted, it cannot be
      removed.
    * The new firmware for the specified bundle level is retrieved from the IBM
      support site or from an FTP server. If no bundle level is specified, but
      an FTP server, all firmware available on the FTP server is retrieved.
      If no bundle level is specified and no FTP server, the already locally
      available firmware is used and no additional firmware is retrieved.
    * The specified firmware is installed.
    * The newly installed firmware is activated, which includes rebooting the SE
      of this CPC.

    If an error occurrs when installing the upgrades for the components of
    the new bundle, any components that were successfully installed are
    rolled back.

    If an error occurs after the firmware is accepted, the firmware remains
    accepted.

    Note that it is not possible to downgrade the SE firmware with this
    operation.

    If the SE firmware is already at the requested bundle level, nothing is
    changed and the command succeeds.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_cpc_upgrade(cmd_ctx, cpc, options))


@cpc_group.command('install-firmware',
                   options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.option('--bundle-level', '-b', type=str, required=False,
              help="Selects the updates to be installed to be those in a "
              "specific SE bundle (e.g. 'S71'). Disruptive updates will fail.")
@click.option('--ec-levels', '-e', type=str, required=False,
              help="Selects the updates to be installed to be specific EC "
              "levels, as a list in YAML Flow Collection style. "
              "The list items are strings of the form 'EC.MCL' where EC is "
              "the EC number of the EC stream, and MCL is the MCL number "
              "within the EC stream. "
              "Example: --ec-levels \"[P30719.015, P30730.007]\"")
@click.option('--all-concurrent', '-c', is_flag=True, required=False,
              help="Selects the updates to be installed to be all "
              "concurrent (= non-disruptive) updates that are locally "
              "available on the SE.")
@click.option('--all', '-a', is_flag=True, required=False,
              help="Selects the updates to be installed to be all updates "
              "that are locally available on the SE, including disruptive "
              "updates.")
@click.option('--install-disruptive', is_flag=True, required=False,
              help="Install any disruptive updates that are encountered. "
              "Only allowed with --ec-levels. "
              "Default: Fail when encountering disruptive updates.")
@click.option('--timeout', '-T', type=int, required=False, default=1200,
              help='Timeout (in seconds) when waiting for the firmware '
              'installation to be complete. Default: 1200.')
@click.pass_obj
def cpc_install_firmware(cmd_ctx, cpc, **options):
    """
    Install retrieved firmware updates on the Support Element (SE) of a CPC.

    This is done by performing the "CPC Install and Activate" operation
    which performs the following steps:

    \b
    * The specified updates are installed.
    * If all updates are installed successfully, they are activated, which
      includes rebooting the SE.

    The updates to be installed must already be available on the SE; they are
    *not* automatically downloaded from the IBM support site or from an FTP
    server.

    If an error occurs when installing the updates, any updates that were
    successfully installed are rolled back.

    If --bundle-level is specified and the SE firmware is already at the
    requested bundle level, nothing is changed and the command succeeds. For
    the other options to select firmware, it is not currently possible to
    distinguish failure from no need to upgrade, so a failure is reported.

    Notes:

    \b
    * This operation does *not* perform a backup, an accept of previously
      activated updates, or an accept of the newly installed updates.
    * This operation does not require that previously activated updates are
      first accepted before invoking this operation.
    * It is not possible to downgrade the SE firmware with this operation.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(
        lambda: cmd_cpc_install_firmware(cmd_ctx, cpc, options))


@cpc_group.command('delete-uninstalled-firmware',
                   options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.option('--ec-levels', '-e', type=str, required=False,
              help="Selects the updates to be deleted to be specific EC "
              "levels, as a list in YAML Flow Collection style. "
              "The list items are strings of the form 'EC.MCL' where EC is "
              "the EC number of the EC stream, and MCL is the MCL number "
              "within the EC stream. "
              "Example: --ec-levels \"[P30719.015, P30730.007]\"")
@click.option('--all', '-a', is_flag=True, required=False,
              help="Selects the updates to be deleted to be all retrieved but "
              "uninstalled updates.")
@click.option('--timeout', '-T', type=int, required=False, default=1200,
              help='Timeout (in seconds) when waiting for the deletion of '
              'updates to be complete. Default: 1200.')
@click.pass_obj
def cpc_delete_uninstalled_firmware(cmd_ctx, cpc, **options):
    """
    Delete retrieved but uninstalled firmware updates on the Support Element
    (SE) of a CPC.

    This is done by performing the "CPC Delete Retrieved Internal Code"
    operation which performs the following steps:

    * The specified updates are deleted from the SE.

    Notes:

    \b
    * This operation does *not* perform a backup.
    * It is not possible to delete installed updates with this command.
    * It is not possible to downgrade the SE firmware with this command.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(
        lambda: cmd_cpc_delete_uninstalled_firmware(cmd_ctx, cpc, options))


@cpc_group.group('hw-message', options_metavar=COMMAND_OPTIONS_METAVAR)
def cpc_hwmessage_group():
    """
    Command group for managing Hardware Messages for a CPC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """


@cpc_hwmessage_group.command('list', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.option('--all', is_flag=True, required=False,
              help='Show all properties.')
@click.option('--service-supported', type=bool, required=False, default=None,
              help='Filter that narrows the list of returned messages to '
              'those with the specified service-supported value. Note: '
              'Dependent on the number of messages, this may take a while.')
@click.option('--begin', type=str, metavar='TIMESTAMP', required=False,
              help='Filter that narrows the list of returned messages to '
              'those created on or after the specified point in time. '
              'Valid formats are HMC timestamp values and most known formats '
              'to represent date and time. An omitted timezone defaults to '
              'UTC. Other omitted fields default to their earliest possible '
              'values. Ambiguous 3-integer dates (e.g. 01/05/09) are '
              'interpreted as M/D/Y.')
@click.option('--end', type=str, metavar='TIMESTAMP', required=False,
              help='Filter that narrows the list of returned messages to '
              'those created on or before the specified point in time. '
              'Valid formats are HMC timestamp values and most known formats '
              'to represent date and time. An omitted timezone defaults to '
              'UTC. Other omitted fields default to their latest possible '
              'values. Ambiguous 3-integer dates (e.g. 01/05/09) are '
              'interpreted as M/D/Y.')
@click.pass_obj
def cpc_hwmessage_list(cmd_ctx, cpc, **options):
    """
    List the Hardware Messages for a CPC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_cpc_hwmessage_list(cmd_ctx, cpc, options))


@cpc_hwmessage_group.command('show', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('message_id', type=str, metavar='MESSAGE_ID')
@click.pass_obj
def cpc_hwmessage_show(cmd_ctx, cpc, message_id):
    """
    Show a Hardware Message for a CPC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_cpc_hwmessage_show(
        cmd_ctx, cpc, message_id))


@cpc_hwmessage_group.command('delete', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('message_id', type=str, metavar='MESSAGE_ID')
@click.pass_obj
def cpc_hwmessage_delete(cmd_ctx, cpc, message_id):
    """
    Delete a Hardware Message for a CPC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_cpc_hwmessage_delete(
        cmd_ctx, cpc, message_id))


@cpc_hwmessage_group.command(
    'request-service', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('message_id', type=str, metavar='MESSAGE_ID')
@click.option('--customer-name', type=str, required=False,
              help='Name of the person that can be contacted about the '
              'problem. Optional, default is the customer name registered with '
              'IBM for the machine.')
@click.option('--customer-phone', type=str, required=False,
              help='Telephone number of the person that can be contacted '
              'about the problem. Optional, default is the customer phone '
              'registered with IBM for the machine.')
@click.pass_obj
def cpc_hwmessage_request_service(cmd_ctx, cpc, message_id, **options):
    """
    Request service from IBM for a Hardware Message for a CPC and delete the
    hardware message.

    The hardware message's "service-supported" property must be True.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_cpc_hwmessage_request_service(
        cmd_ctx, cpc, message_id, options))


@cpc_hwmessage_group.command(
    'get-service-info', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('message_id', type=str, metavar='MESSAGE_ID')
@click.option('--delete', is_flag=True, required=False,
              help='Controls whether the hardware message will be deleted '
              'upon successful completion of the operation.')
@click.pass_obj
def cpc_hwmessage_get_service_info(cmd_ctx, cpc, message_id, **options):
    """
    Get problem information and a telephone number for requesting service from
    IBM for a hardware message for a CPC and optionally delete the hardware
    message.

    The hardware message's "service-supported" property must be True.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_cpc_hwmessage_get_service_info(
        cmd_ctx, cpc, message_id, options))


@cpc_hwmessage_group.command(
    'decline-service', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('message_id', type=str, metavar='MESSAGE_ID')
@click.pass_obj
def cpc_hwmessage_decline_service(cmd_ctx, cpc, message_id):
    """
    Decline service from IBM for a Hardware Message for a CPC and delete
    the hardware message.

    The hardware message's "service-supported" property must be True.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_cpc_hwmessage_decline_service(
        cmd_ctx, cpc, message_id))


def cmd_cpc_list(cmd_ctx, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)

    filter_args = build_filter_args(cmd_ctx, options['filter'])
    try:
        cpcs = client.cpcs.list(filter_args=filter_args)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    if options['type']:
        click.echo("The --type option is deprecated and type information "
                   "is now always shown.")
    if options['mach']:
        click.echo("The --mach option is deprecated and machine information "
                   "is now always shown.")

    show_list = [
        'name',
    ]
    if not options['names_only']:
        show_list.extend([
            'status',
            'dpm-enabled',
            'se-version',
            'machine-type',
            'machine-model',
            'machine-serial-number',
            'description',
        ])
    if options['uri']:
        show_list.extend([
            'object-uri',
        ])

    sort_props = build_sort_props(cmd_ctx, options['sort'], default=['name'])
    try:
        print_resources(cmd_ctx, cpcs, cmd_ctx.output_format, show_list,
                        all=options['all'], sort_props=sort_props)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)


def cmd_cpc_show(cmd_ctx, cpc_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)

    try:
        cpc.pull_full_properties()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    properties = dict(cpc.properties)

    # Hide some long or deeply nested properties in table output formats.
    if not options['all'] and cmd_ctx.output_format in TABLE_FORMATS:
        hide_property(properties, 'auto-start-list')
        hide_property(properties, 'available-features-list')
        hide_property(properties, 'cpc-power-saving-state')
        hide_property(properties, 'ec-mcl-description')
        hide_property(properties, 'network1-ipv6-info')
        hide_property(properties, 'network2-ipv6-info')
        hide_property(properties, 'stp-configuration')

    print_properties(cmd_ctx, properties, cmd_ctx.output_format)


def cmd_cpc_update(cmd_ctx, cpc_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)

    name_map = {
        'next-activation-profile': 'next-activation-profile-name',
        'processor-time-slice': None,
        'wait-ends-slice': None,
        'no-wait-ends-slice': None,  # sets 'wait-ends-slice' option
        'acceptable-status': None,
    }
    org_options = original_options(options)
    properties = options_to_properties(org_options, name_map)

    time_slice = org_options['processor-time-slice']
    if time_slice is None:
        # 'processor-running-time*' properties not changed
        pass
    elif time_slice < 0:
        raise click_exception("Value for processor-time-slice option must "
                              "be >= 0", cmd_ctx.error_format)
    elif time_slice == 0:
        properties['processor-running-time-type'] = 'system-determined'
    else:  # time_slice > 0
        properties['processor-running-time-type'] = 'user-determined'
        properties['processor-running-time'] = time_slice

    if org_options['wait-ends-slice'] is not None:
        properties['does-wait-state-end-time-slice'] = \
            org_options['wait-ends-slice']

    if org_options['acceptable-status'] is not None:
        status_list = org_options['acceptable-status'].split(',')
        status_list = [item for item in status_list if item]
        properties['acceptable-status'] = status_list

    if not properties:
        cmd_ctx.spinner.stop()
        click.echo("No properties specified for updating CPC '{c}'.".
                   format(c=cpc_name))
        return

    try:
        cpc.update_properties(properties)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()

    # Name changes are not supported for CPCs.
    click.echo(f"CPC '{cpc_name}' has been updated.")


def cmd_cpc_set_power_save(cmd_ctx, cpc_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)

    org_options = original_options(options)
    power_saving = org_options['power-saving']
    cpc.set_power_save(power_saving)
    cmd_ctx.spinner.stop()
    click.echo("The power save settings of CPC '{c}' have been set to {s}.".
               format(c=cpc_name, s=power_saving))


def cmd_cpc_set_power_capping(cmd_ctx, cpc_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)

    org_options = original_options(options)
    power_capping_state = org_options['power-capping-state']
    power_cap_current = None
    if org_options['power-cap-current']:
        power_cap_current = org_options['power-cap-current']
    cpc.set_power_capping(org_options['power-capping-state'], power_cap_current)
    cmd_ctx.spinner.stop()
    click.echo("The power capping settings of CPC '{c}' have been set to {s}.".
               format(c=cpc_name, s=power_capping_state))


def cmd_cpc_get_em_data(cmd_ctx, cpc_name):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)

    energy_props = cpc.get_energy_management_properties()
    print_properties(cmd_ctx, energy_props, cmd_ctx.output_format)


def cmd_dpm_export(cmd_ctx, cpc_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)

    dpm_file = required_option(options['dpm_file'], '--dpm-file')
    dpm_format = options['dpm_format']

    include_unused_adapters = options['include_unused_adapters']

    config_dict = cpc.export_dpm_configuration(include_unused_adapters)
    if not options['exclude_meta_fields']:
        now = datetime.now(timezone.utc)
        config_dict['zhmccli-meta-exported-by'] = cmd_ctx.session.userid
        config_dict['zhmccli-meta-exported-from-cpc-name'] = cpc_name
        config_dict['zhmccli-meta-exported-when'] = f'{now} UTC'
        config_dict['zhmccli-meta-zhmccli-version'] = __version__
        # pylint: disable=no-member
        config_dict['zhmccli-meta-zhmcclient-version'] = zhmcclient.__version__

    try:
        with open(dpm_file, 'w', encoding='utf-8') as fp:
            if dpm_format == 'yaml':
                yaml.dump(config_dict, fp, encoding=None, allow_unicode=True,
                          default_flow_style=False, indent=2,
                          Dumper=yamlloader.ordereddict.CSafeDumper)
            else:
                assert dpm_format == 'json'
                json.dump(config_dict, fp, indent=2, sort_keys=True)
    except OSError as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("Exported DPM configuration of CPC '{c}' into DPM config "
               "file {f} in {ff} format.".
               format(c=cpc_name, f=dpm_file, ff=dpm_format.upper()))
    _dump_config(config_dict, 'Export data summary:')


def cmd_dpm_import(cmd_ctx, cpc_name, options):
    # pylint: disable=missing-function-docstring

    # process options first, without the spinner running
    cmd_ctx.spinner.stop()

    dpm_file = required_option(options['dpm_file'], '--dpm-file')
    dpm_format = options['dpm_format']
    mapping_file = options['mapping_file']
    preserve_uris = options['preserve_uris']
    preserve_wwpns = options['preserve_wwpns']

    try:
        with open(dpm_file, encoding='utf-8') as fp:
            if dpm_format == 'yaml':
                try:
                    config_dict = yaml.safe_load(fp)
                except (yaml.parser.ParserError, yaml.scanner.ScannerError) \
                        as exc:
                    raise click_exception(
                        "Error parsing DPM configuration file {} in YAML "
                        "format: {}".format(dpm_file, exc),
                        cmd_ctx.error_format)
            else:
                assert dpm_format == 'json'
                try:
                    config_dict = json.load(fp)
                except ValueError as exc:
                    raise click_exception(
                        "Error parsing DPM configuration file {} in JSON "
                        "format: {}".format(dpm_file, exc),
                        cmd_ctx.error_format)
    except OSError as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    # we collect information about the preserve flags, adapter mapping, and
    # meta fields as list of lists, to then print via tabulate()
    summary_info = []
    summary_info.append(_fetch_and_handle_preserve_flag(
        config_dict, 'preserve-uris', preserve_uris, dpm_file))
    summary_info.append(_fetch_and_handle_preserve_flag(
        config_dict, 'preserve-wwpns', preserve_wwpns, dpm_file))
    if mapping_file:
        try:
            with open(mapping_file, encoding='utf-8') as fp:
                try:
                    mapping_obj = yaml.safe_load(fp)
                except (yaml.parser.ParserError, yaml.scanner.ScannerError) \
                        as exc:
                    raise click_exception(
                        "Error parsing adapter mapping file {} in YAML "
                        "format: {}".format(dpm_file, exc),
                        cmd_ctx.error_format)
        except OSError as exc:
            raise click_exception(exc, cmd_ctx.error_format)
    else:
        mapping_obj = None

    summary_info.append(_fetch_and_handle_mapping(
        cmd_ctx, config_dict, mapping_obj, dpm_file, mapping_file))
    summary_info.append([])
    summary_info.extend(_fetch_and_drop_meta_fields(config_dict))
    summary_info.append([])
    click.echo(tabulate(summary_info, [], "plain"))

    _dump_config(config_dict, "Import data summary:")
    cmd_ctx.spinner.stop()

    if not options["yes"]:
        message = ("Are you sure you want to import the DPM configuration "
                   "from {f} into CPC {c}, replacing its current DPM "
                   "configuration?").format(c=cpc_name, f=dpm_file)

        if not click.confirm(message):
            return

    cmd_ctx.spinner.start()

    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)
    result = cpc.import_dpm_configuration(config_dict)

    cmd_ctx.spinner.stop()

    if result:
        click.echo("Partially imported DPM configuration from DPM config file "
                   "{f} into CPC '{c}'. The following parts were not restored:"
                   .format(c=cpc_name, f=dpm_file))
        try:
            print_dicts(cmd_ctx, result['output'], cmd_ctx.output_format)
        except zhmcclient.Error as exc:
            raise click_exception(exc, cmd_ctx.error_format)
    else:
        click.echo("Imported DPM configuration from DPM config file {f} into "
                   "CPC '{c}'.".format(c=cpc_name, f=dpm_file))


def _fetch_and_handle_preserve_flag(config_dict, key, options_field, dpm_file):
    """
    Checks whether key is specified as zhmc option.
    If so, the corresponding value is read and stored within config_dict.
    Returns: list with 2 values (field name and field value with its "source")
    """
    flag = f'{key}:'
    if options_field is not None:
        # zhmc takes precedence over configuration file
        config_dict[key] = bool(options_field)
        return [flag, f'{bool(options_field)} from zhmc option']

    if key in config_dict:
        return [flag, f'{config_dict[key]} from {dpm_file}']

    return [flag, 'False from HMC default']


def _fetch_and_handle_mapping(cmd_ctx, config_dict, mapping_obj,
                              dpm_file, mapping_file):
    """
    Checks whether adapter-mapping is specified as zhmc option.
    If so, the mapping file is read and stored within config_dict.
    Returns: list with 2 values (field name and field value with its "source")
    """
    if mapping_obj:
        # zhmc takes precedence over configuration file
        config_dict['adapter-mapping'] = \
            convert_adapter_mapping(cmd_ctx, mapping_obj)
        return ['adapter-mapping:', f'from {mapping_file}']

    if 'adapter-mapping' in config_dict:
        return ['adapter-mapping:', f'from {dpm_file}']

    return ['adapter-mapping:', '1 to 1 from HMC default']


def _fetch_and_drop_meta_fields(config_dict):
    """
    Prints all dict keys starting with 'zhmccli-' and drops them from the dict.
    Returns: list with value pairs: field name and field value
    """
    summary = []
    for k in sorted(config_dict):
        if k.startswith('zhmccli-'):
            summary.append([f'{k}:', config_dict[k]])
            config_dict.pop(k)
    return summary


def convert_adapter_mapping(cmd_ctx, mapping_obj):
    """
    Convert the adapter mapping from the format used in the adapter mapping file
    to the format needed by the adapter-mapping property in the Import DPM
    Configuration operation.
    """
    try:
        validate(mapping_obj, MAPPING_SCHEMA, "adapter mapping file")
    except ValueError as exc:
        raise click_exception(exc, cmd_ctx.error_format)
    ret_mapping = []
    mapping_list = mapping_obj['adapter-mapping']
    for mapping in mapping_list:
        ret_item = {}
        ret_item['new-adapter-id'] = mapping['new-adapter-id']
        ret_item['old-adapter-id'] = mapping['old-adapter-id']
        ret_mapping.append(ret_item)
    return ret_mapping


def _dump_config(config_dict, message):
    """
    Summarizes the key elements of a dict-based DPM configuration
    """
    counts = []
    values = []
    for k in config_dict:
        v = config_dict[k]
        if isinstance(v, list):
            counts.append((f'{len(v):>3}', k))
        else:
            if isinstance(v, (bool, str)):
                values.append((f'{k}:', v))
    click.echo(message)
    click.echo(tabulate(counts, [], "plain"))
    click.echo(tabulate(values, [], "plain"))


def get_auto_start_list(cpc):
    """
    Helper function that converts the 'auto-start-list' property of a CPC
    to a list suitable for the zhmcclient.Cpc.set_auto_start_list() method.

    Returns:
        None - if the CPC is in classic mode
        list, with items that are one of:
          - tuple(partition, post_start_delay)
          - tuple(partition_list, name, description, post_start_delay)
    """
    auto_start_list = cpc.prop('auto-start-list', None)
    if auto_start_list is None:
        # CPC is in classic mode
        return None

    as_list = []
    for auto_start_item in auto_start_list:
        if auto_start_item['type'] == 'partition':
            # item is a partition
            uri = auto_start_item['partition-uri']
            delay = auto_start_item['post-start-delay']
            partition = cpc.partitions.resource_object(uri)
            as_item = (partition, delay)
            as_list.append(as_item)
        if auto_start_item['type'] == 'partition-group':
            # item is a partition group
            name = auto_start_item['name']
            description = auto_start_item['description']
            delay = auto_start_item['post-start-delay']
            partitions = []
            for uri in auto_start_item['partition-uris']:
                partition = cpc.partitions.resource_object(uri)
                partitions.append(partition)
            as_item = (partitions, name, description, delay)
            as_list.append(as_item)
    return as_list


def auto_start_table_str(as_list, output_format):
    """
    Return a string with the auto-start list table in the specified output
    format.
    """
    headers = ['Partition/Group', 'Post start delay', 'Partitions in group',
               'Group description']
    table = []
    for as_item in as_list:
        if isinstance(as_item[0], zhmcclient.Partition):
            # item is a partition
            partition, delay = as_item
            row = [partition.name, delay]
            table.append(row)
        else:
            # item is a partition group
            partitions, name, description, delay = as_item
            partition_names = ', '.join([p.name for p in partitions])
            row = [name, delay, partition_names, description]
            table.append(row)
    table_str = tabulate(table, headers, tablefmt=output_format)
    return table_str


def cmd_cpc_autostart_show(cmd_ctx, cpc_name):
    # pylint: disable=missing-function-docstring
    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)

    as_list = get_auto_start_list(cpc)

    if as_list is None:
        # The HMC WS API book documents that only CPCs in DPM mode have
        # the 'auto-start-list' property.
        cmd_ctx.spinner.stop()
        click.echo("CPC '{c}' is in classic mode and has no auto-start list.".
                   format(c=cpc_name))
        return

    table_str = auto_start_table_str(as_list, cmd_ctx.output_format)
    cmd_ctx.spinner.stop()
    click.echo(table_str)


def cmd_cpc_autostart_add(cmd_ctx, cpc_name, partitions_delay, options):
    # pylint: disable=missing-function-docstring
    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)
    partition_names, delay = partitions_delay

    # pylint: disable=import-outside-toplevel,cyclic-import
    from ._cmd_partition import find_partition

    as_list = get_auto_start_list(cpc)

    if as_list is None:
        # The HMC WS API book documents that only CPCs in DPM mode have
        # the 'auto-start-list' property.
        cmd_ctx.spinner.stop()
        click.echo("CPC '{c}' is in classic mode and has no auto-start list.".
                   format(c=cpc_name))
        return

    group_name = options['group']
    if group_name:
        # A partition group is added (with one or more partitions)
        partition_names = partition_names.split(',')
        partitions = []
        for partition_name in partition_names:
            partition = find_partition(cmd_ctx, client, cpc_name,
                                       partition_name)
            partitions.append(partition)
        description = options['description']
        new_as_item = (partitions, group_name, description, delay)
    else:
        # A partition is added
        partition_name = partition_names
        partition = find_partition(cmd_ctx, client, cpc_name, partition_name)
        new_as_item = (partition, delay)

    # TODO: Add support for --before and --after
    as_list.append(new_as_item)
    try:
        cpc.set_auto_start_list(as_list)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    table_str = auto_start_table_str(as_list, cmd_ctx.output_format)
    cmd_ctx.spinner.stop()
    click.echo(table_str)


def cmd_cpc_autostart_delete(cmd_ctx, cpc_name, partition_or_group):
    # pylint: disable=missing-function-docstring
    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)

    as_list = get_auto_start_list(cpc)

    if as_list is None:
        # The HMC WS API book documents that only CPCs in DPM mode have
        # the 'auto-start-list' property.
        cmd_ctx.spinner.stop()
        click.echo("CPC '{c}' is in classic mode and has no auto-start list.".
                   format(c=cpc_name))
        return

    as_item_idx = None
    for i, as_item in enumerate(as_list):
        if isinstance(as_item[0], zhmcclient.Partition):
            # item is a partition
            partition = as_item[0]
            if partition.name == partition_or_group:
                as_item_idx = i
                break
        else:
            # item is a partition group
            name = as_item[1]
            if name == partition_or_group:
                as_item_idx = i
                break

    if as_item_idx is None:
        raise click_exception(
            "Could not find partition or group '{p}' in CPC '{c}'.'".
            format(p=partition_or_group, c=cpc_name),
            cmd_ctx.error_format)

    del as_list[as_item_idx]
    try:
        cpc.set_auto_start_list(as_list)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    table_str = auto_start_table_str(as_list, cmd_ctx.output_format)
    cmd_ctx.spinner.stop()
    click.echo(table_str)


def cmd_cpc_autostart_clear(cmd_ctx, cpc_name):
    # pylint: disable=missing-function-docstring
    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)

    as_list = []
    try:
        cpc.set_auto_start_list(as_list)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("Auto-start list for CPC '{c}' has been cleared.".
               format(c=cpc_name))


def cmd_cpc_list_api_features(cmd_ctx, cpc_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)

    name = options['name']
    try:
        features = cpc.list_api_features(name)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    print_list(cmd_ctx, features, cmd_ctx.output_format)


def cmd_cpc_list_firmware(cmd_ctx, cpc_name, options):
    # pylint: disable=missing-function-docstring,unused-argument

    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)
    cpc.pull_properties('ec-mcl-description')
    ec_mcl = cpc.properties['ec-mcl-description']

    firmware_list = convert_ec_mcl_description(ec_mcl)

    # define order of columns in output table
    show_list = [
        'ec-number',
        'description',
        'retrieved',
        'installable-conc',
        'activated',
        'accepted',
        'removable-conc',
    ]

    print_dicts(cmd_ctx, firmware_list, cmd_ctx.output_format,
                show_list=show_list)


def cmd_cpc_upgrade(cmd_ctx, cpc_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)
    console = client.consoles.console

    bundle_level = options['bundle_level']
    accept_firmware = options['accept_firmware']
    timeout = options['timeout']

    ftp_host = options['ftp_host']
    ftp_user = options['ftp_user']
    ftp_password = options['ftp_password']
    if ftp_host and ftp_password == '-':  # nosec: B105
        ftp_password = prompt_ftp_password(cmd_ctx, ftp_host, ftp_user)

    ec_mcl = console.prop('ec-mcl-description')
    hmc_bundle_level = ec_mcl.get('bundle-level', None)
    if hmc_bundle_level is None:
        hmc_version = console.prop('version')
        raise click_exception(
            "HMC version {v} does not support firmware upgrade through "
            "the Web Services API".format(v=hmc_version),
            cmd_ctx.error_format)

    level_str = get_level_str(bundle_level, ftp_host)
    click.echo("Upgrading the SE of CPC {c} to {lvl}, and waiting for "
               "completion (timeout: {t} s)".
               format(c=cpc_name, lvl=level_str, t=timeout))

    kwargs = dict(
        bundle_level=bundle_level,
        accept_firmware=accept_firmware,
        wait_for_completion=True,
        operation_timeout=timeout)
    if ftp_host:
        kwargs['ftp_host'] = ftp_host
        kwargs['ftp_protocol'] = options['ftp_protocol']
        kwargs['ftp_user'] = ftp_user
        kwargs['ftp_password'] = ftp_password
        kwargs['ftp_directory'] = options['ftp_directory']

    try:
        cpc.single_step_install(**kwargs)
    except zhmcclient.HTTPError as exc:
        if exc.http_status == 400 and exc.reason == 356:
            # HMC was already at that bundle level
            cmd_ctx.spinner.stop()
            click.echo("The SE of CPC {c} was already at {lvl} and did not "
                       "need to be upgraded".
                       format(c=cpc_name, lvl=level_str))
            return
        raise click_exception(exc, cmd_ctx.error_format)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("The SE of CPC {c} has been upgraded to {lvl}".
               format(c=cpc_name, lvl=level_str))


def cmd_cpc_install_firmware(cmd_ctx, cpc_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)

    bundle_level = options['bundle_level']
    ec_levels = options['ec_levels']
    all_ = options['all']
    concurrent = options['all_concurrent']
    install_disruptive = options['install_disruptive']
    timeout = options['timeout']

    num_mcl_options = bool(bundle_level) + bool(ec_levels) + all_ + concurrent
    if num_mcl_options != 1:
        raise click_exception(
            "Exactly one option for specifying the firmware to be installed "
            "must be specified, but there were {}".format(num_mcl_options),
            cmd_ctx.error_format)

    if bundle_level and install_disruptive:
        raise click_exception(
            "--install-disruptive and --bundle-level cannot both be specified",
            cmd_ctx.error_format)

    if concurrent and install_disruptive:
        raise click_exception(
            "--install-disruptive and --concurrent cannot both be specified",
            cmd_ctx.error_format)

    if all_ and install_disruptive:
        raise click_exception(
            "--install-disruptive and --all cannot both be specified",
            cmd_ctx.error_format)

    if ec_levels:
        ec_levels_parm = parse_ec_levels(cmd_ctx, '--ec-levels', ec_levels)

    if all_:
        bundle_level = None
        ec_levels_parm = None
        install_disruptive = True

    if concurrent:
        bundle_level = None
        ec_levels_parm = None
        install_disruptive = False

    level_str, dis_str = get_mcl_str(
        bundle_level, ec_levels, all_, concurrent, install_disruptive)
    click.echo("Upgrading the SE of CPC {c} to {lvl}{dis}, and waiting for "
               "completion (timeout: {t} s)".
               format(c=cpc_name, lvl=level_str, dis=dis_str, t=timeout))

    kwargs = dict(
        wait_for_completion=True,
        operation_timeout=timeout,
    )
    if install_disruptive:
        kwargs['install_disruptive'] = True
    if bundle_level:
        kwargs['bundle_level'] = bundle_level
    if ec_levels:
        kwargs['ec_levels'] = ec_levels_parm

    try:
        cpc.install_and_activate(**kwargs)
    except zhmcclient.HTTPError as exc:
        if bundle_level and exc.http_status == 500 and exc.reason == 263:
            # --bundle-level was specified, and the SE was already at the
            # requested bundle level.
            cmd_ctx.spinner.stop()
            click.echo("The SE of CPC {c} was already at {lvl} and did not "
                       "need to be upgraded".
                       format(c=cpc_name, lvl=level_str))
            return
        if ec_levels and exc.http_status == 400 and exc.reason == 379:
            # --ec-levels was specified, but we cannot distinguish failure from
            # the case where the SE was already at the requested EC levels.
            # TODO: Get EC/MCL info and figure out which case it is.
            cmd_ctx.spinner.stop()
            raise click_exception(
                "The SE of CPC {c} either was already at {lvl}, or above "
                "(cannot downgrade), or the firmware is not available on the "
                "SE: HTTPError: {exc}".
                format(c=cpc_name, lvl=level_str, exc=exc),
                cmd_ctx.error_format)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("The SE of CPC {c} has been upgraded to {lvl}".
               format(c=cpc_name, lvl=level_str))


def cmd_cpc_delete_uninstalled_firmware(cmd_ctx, cpc_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)

    ec_levels = options['ec_levels']
    all_ = options['all']
    timeout = options['timeout']

    num_mcl_options = bool(ec_levels) + all_
    if num_mcl_options != 1:
        raise click_exception(
            "Exactly one option for specifying the firmware to be deleted must "
            "be specified, but there were {}".format(num_mcl_options),
            cmd_ctx.error_format)

    if ec_levels:
        ec_levels_parm = parse_ec_levels(cmd_ctx, '--ec-levels', ec_levels)

    if all_:
        ec_levels_parm = None

    level_str, _ = get_mcl_str(None, ec_levels, all_, None, False)
    click.echo("Deleting {lvl} from the SE of CPC {c}, and waiting for "
               "completion (timeout: {t} s)".
               format(c=cpc_name, lvl=level_str, t=timeout))

    kwargs = dict(
        wait_for_completion=True,
        operation_timeout=timeout,
    )
    if ec_levels:
        kwargs['ec_levels'] = ec_levels_parm

    try:
        cpc.delete_retrieved_internal_code(**kwargs)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    level_str = level_str[0].upper() + level_str[1:]
    click.echo("{lvl} have been deleted from the SE of CPC {c}".
               format(c=cpc_name, lvl=level_str))


def cmd_cpc_hwmessage_list(cmd_ctx, cpc_name, options):
    # pylint: disable=missing-function-docstring
    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)

    begin_time = options['begin']
    end_time = options['end']
    service_supported = options['service_supported']

    if begin_time is not None:
        try:
            begin_time = parse_timestamp(begin_time, TIMESTAMP_BEGIN_DEFAULT)
        except (ValueError, OverflowError) as exc:
            raise click_exception(
                f"Invalid begin time: {exc}", cmd_ctx.error_format)
    if end_time is not None:
        try:
            end_time = parse_timestamp(end_time, TIMESTAMP_END_DEFAULT)
        except (ValueError, OverflowError) as exc:
            raise click_exception(
                f"Invalid end time: {exc}", cmd_ctx.error_format)

    filter_args = {}
    if service_supported is not None:
        filter_args['service-supported'] = service_supported
    try:
        messages = cpc.hw_messages.list(
            filter_args=filter_args,
            begin_time=begin_time,
            end_time=end_time,
        )
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    show_list = [
        'timestamp-utc',
        'message-id',
        'text',
    ]

    ts_additions = {}
    msgid_additions = {}
    for message in messages:
        hmc_ts = message.prop('timestamp')
        dt = zhmcclient.datetime_from_timestamp(hmc_ts)
        ts_additions[message.uri] = dt.strftime('%Y-%m-%d %H:%M:%S.%f')
        msgid_additions[message.uri] = message.name
    additions = {
        'timestamp-utc': ts_additions,
        'message-id': msgid_additions,
    }

    try:
        print_resources(cmd_ctx, messages, cmd_ctx.output_format,
                        show_list, additions=additions, all=options['all'])
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)


def cmd_cpc_hwmessage_show(cmd_ctx, cpc_name, message_id):
    # pylint: disable=missing-function-docstring
    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)
    message = find_cpc_hwmessage(cmd_ctx, cpc, message_id)
    message.pull_full_properties()

    message.update_properties_local({'message-id': message.name})
    message.update_properties_local({'parent-name': cpc.name})
    hmc_ts = message.prop('timestamp')
    dt = zhmcclient.datetime_from_timestamp(hmc_ts)
    dt_str = dt.strftime('%Y-%m-%d %H:%M:%S.%f')
    message.update_properties_local({'timestamp-utc': dt_str})

    cmd_ctx.spinner.stop()
    print_properties(cmd_ctx, message.properties, cmd_ctx.output_format)


def cmd_cpc_hwmessage_delete(cmd_ctx, cpc_name, message_id):
    # pylint: disable=missing-function-docstring
    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)
    message = find_cpc_hwmessage(cmd_ctx, cpc, message_id)

    try:
        message.delete()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo(f"Hardware Message '{message.name}' has been deleted.")


def cmd_cpc_hwmessage_request_service(cmd_ctx, cpc_name, message_id, options):
    # pylint: disable=missing-function-docstring
    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)
    message = find_cpc_hwmessage(cmd_ctx, cpc, message_id)

    try:
        message.request_service(**options)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo(f"IBM service for Hardware Message '{message.name}' has been "
               "requested and the message has been deleted.")


def cmd_cpc_hwmessage_get_service_info(cmd_ctx, cpc_name, message_id, options):
    # pylint: disable=missing-function-docstring
    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)
    message = find_cpc_hwmessage(cmd_ctx, cpc, message_id)
    delete = options['delete']

    try:
        info = message.get_service_information(delete=delete)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    print_properties(cmd_ctx, info, cmd_ctx.output_format)
    if delete:
        click.echo(f"Hardware Message '{message.name}' has been deleted.")


def cmd_cpc_hwmessage_decline_service(cmd_ctx, cpc_name, message_id):
    # pylint: disable=missing-function-docstring
    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)
    message = find_cpc_hwmessage(cmd_ctx, cpc, message_id)

    try:
        message.decline_service()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo(f"IBM service for Hardware Message '{message.name}' has been "
               "declined and the message has been deleted.")
